/* This example mimic PnetCDF APIs to write a 1D array of float type */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <mpi.h>
#include <damsel.h>

#define DIM_LEN 15

#define CHECK_ERR(e) {if ((e.id!=DMSL_OK.id)) { printf("Error at line=%d msg=%s\n", __LINE__,DMSLerror_str(e)); goto fn_exit;}}

static damsel_err_t mimic_ncmpi_file_create(MPI_Comm comm, char *filename, char *cmode, damsel_model *model);
static damsel_err_t mimic_ncmpi_def_dim(damsel_model model, damsel_container *container, int dim_len, MPI_Comm comm);
static damsel_err_t mimic_ncmpi_def_var(damsel_model model, damsel_container container, char *name, damsel_data_type type, void *buf, damsel_handle *tag_handle);
static damsel_err_t mimic_ncmpi_enddef(damsel_model model, damsel_container container, damsel_handle tag_handle);
static damsel_err_t mimic_ncmpi_iput(damsel_model model, damsel_request_t *req);
static damsel_err_t mimic_ncmpi_wait(int num, damsel_request_t *req, damsel_status_t *status);
static damsel_err_t mimic_ncmpi_close(damsel_model model, damsel_container container);

/*----< main() >--------------------------------------------------------------*/
int main(int argc, char **argv) {
    int i, rank, nprocs;
    float *buf;
    damsel_err_t err;

#ifndef DAMSEL_SINGLETHREAD
    int res;
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
    assert(res == MPI_THREAD_MULTIPLE);
#else
    MPI_Init(&argc, &argv);
#endif

    /* initialize Damsel environment */
    damsel_library lib;
    err = DMSLlib_init(&lib);
    CHECK_ERR(err)

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (argc != 2) {
        if (!rank) printf("Usage: %s filename\n",argv[0]);
        goto fn_exit;
    }

    /* create a new file */
    damsel_model model;
    err = mimic_ncmpi_file_create(MPI_COMM_WORLD, argv[1], "WRITE", &model);
    CHECK_ERR(err)

    /* define a dimension of size DIM_LEN (global size) */
    damsel_container container;
    err = mimic_ncmpi_def_dim(model, &container, DIM_LEN, MPI_COMM_WORLD);
    CHECK_ERR(err)

    /* initialize local array of size local_len */
    int local_len = DIM_LEN / nprocs;
    if (rank < DIM_LEN % nprocs) local_len++;
    buf = (float*) malloc(local_len * sizeof(float));
    for (i=0; i<local_len; i++) buf[i] = 1.0 * rank;

    /* define a global variable, a 1D array of float type */
    damsel_handle tag_handle = 123; /* what value should this be? */
    err = mimic_ncmpi_def_var(model, container, "some_array",
                            DAMSEL_DATA_TYPE_FLOAT, buf, &tag_handle);
    CHECK_ERR(err)

    /* end of define */
    err = mimic_ncmpi_enddef(model, container, tag_handle);
    CHECK_ERR(err)

    /* Now we can tell Damsel to start writing the data */
    damsel_request_t req;
    err = mimic_ncmpi_iput(model, &req);
    CHECK_ERR(err)

    /* Other computations can be performed here, i.e. between the post of
     * asynchronous transfer and wait, as long as the registered data buffer,
     * buf, is not touched */

    damsel_status_t  status;
    err = mimic_ncmpi_wait(1, &req, &status);
    CHECK_ERR(err)

    err = mimic_ncmpi_close(model, container);

fn_exit:
    /* exit Damsel environment */
    if (lib != DAMSEL_LIBRARY_INVALID) {
        err = DMSLlib_finalize(lib);
        CHECK_ERR(err)
    }

    MPI_Finalize();
    return 0;
}

/*----< mimic_ncmpi_file_create() >-------------------------------------------*/
static
damsel_err_t mimic_ncmpi_file_create(MPI_Comm      comm,
                                     char         *filename,
                                     char         *cmode,
                                     damsel_model *model)
{
    damsel_err_t err;

    /* create a model, a in-memory store */
    err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, model);
    if (err.id != DMSL_OK.id) return err;

    /* make sure the file does not exist, ignore error if any */
    unlink(filename);

    /* set the file access mode */
    damsel_trait_bundle tb;
    if (err.id != DMSL_OK.id) return err;
    err = DMSLtrait_create(&tb);
    if (err.id != DMSL_OK.id) return err;

    if (!strcasecmp(cmode, "WRITE"))
        err = DMSLtrait_put(tb, "mode", "WRITE");
    else
        err = DMSLtrait_put(tb, "mode", "READ");
    if (err.id != DMSL_OK.id) return err;

    /* register the filename and access mode to model (in-memory store) */
    err = DMSLmodel_attach(*model, filename, MPI_COMM_WORLD, tb);
    if (err.id != DMSL_OK.id) return err;

    err = DMSLtrait_release(tb);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_def_dim() >-----------------------------------------------*/
static
damsel_err_t mimic_ncmpi_def_dim(damsel_model      model,
                                 damsel_container *container,
                                 int               dim_len,
                                 MPI_Comm          comm)
{
    int rank, nprocs, start, count;
    damsel_err_t err;

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &nprocs);

    count = dim_len / nprocs;
    start = count * rank;
    if (rank < dim_len % nprocs) {
        start += rank;
        count++;
    }
    else {
        start += dim_len % nprocs;
    }

    damsel_handle start_handle = start;
    err = DMSLcontainer_create_sequence(model, start_handle, count, 1,
                                        container);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_def_var() >-----------------------------------------------*/
static
damsel_err_t mimic_ncmpi_def_var(damsel_model      model,
                                 damsel_container  container,
                                 char             *name,
                                 damsel_data_type  type,
                                 void             *buf,
                                 damsel_handle    *tag_handle)
{
    damsel_err_t err;
    err = DMSLtag_define(model, tag_handle, type, name);
    if (err.id != DMSL_OK.id) return err;

    /* Tag the container and register I/O buffer (buf) */
    err = DMSLmodel_map_tag(buf, container, tag_handle);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_enddef() >------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_enddef(damsel_model     model,
                                damsel_container container,
                                damsel_handle    tag_handle)
{
    damsel_err_t err;

    /* invent/map file-side handle for each container */
    err = DMSLmodel_map_handles_inventing_file_handles(container);
    if (err.id != DMSL_OK.id) return err;

    /* invent/map a single file-side handle for each tag */ 
    err = DMSLmodel_map_single_handle_inventing_file_handle(model, tag_handle);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_iput() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_iput(damsel_model      model,
                              damsel_request_t *req)
{
    damsel_err_t err;
    err = DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, req);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_wait() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_wait(int               num,
                              damsel_request_t *req,    /* [num] */
                              damsel_status_t  *status) /* [num] */
{
    int i;
    damsel_err_t err;
    /* wait for the I/O transfer to complete */
    for (i=0; i<num; i++) {
        err = DMSLmodel_wait(req[i], &status[i]);
        if (err.id != DMSL_OK.id) return err;
    }
    return DMSL_OK;
}

/*----< mimic_ncmpi_close() >-------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_close(damsel_model     model,
                               damsel_container container)
{
    damsel_err_t err;
    /* close in-memory store */
    err = DMSLmodel_close(model);
    if (err.id != DMSL_OK.id) return err;

    /* close container handle */
    err = DMSLcontainer_close(container);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*
Compile:
mpicc -c -g -I/path/to/Damsel/include array1d.c

Link:
mpicc -o array1d array1d.o -L/path/to/Damsel/lib -L/path/to/HDF5/lib -L/path/to/PnetCDF/lib -ldamsel -lhdf5 -lhdf5_hl -lm -lz -lpnetcdf

Run: (on 2 process in parallel)
mpiexec -n 2 array1d output/dmsl.h5

Output: (from command h5dump)
h5dump output/dmsl.h5

HDF5 "output/dmsl.h5" {
GROUP "/" {
   DATASET "damsel-collections" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "type";
         H5T_STD_U64LE "collection_handle";
         H5T_STD_U64LE "container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "damsel-entity-connectivity" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I32LE "entity_type";
         H5T_STD_I32LE "vertices_per_entity";
         H5T_STD_U64LE "vertex_container_handle";
         H5T_STD_U64LE "entity_container_handle";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( H5S_UNLIMITED ) }
      DATA {
      }
   }
   DATASET "damsel-handle-allocator" {
      DATATYPE  H5T_STD_I64LE
      DATASPACE  SIMPLE { ( 1 ) / ( 1 ) }
      DATA {
      (0): 2
      }
   }
   DATASET "damsel-seq-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_U64LE "base";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "step";
      }
      DATASPACE  SIMPLE { ( 4 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            8589939602,
            8589939594,
            8,
            1
         },
      (1): {
            8589939593,
            8589939592,
            1,
            1
         },
      (2): {
            12884906897,
            12884906890,
            7,
            1
         },
      (3): {
            12884906889,
            12884906888,
            1,
            1
         }
      }
   }
   DATASET "damsel-tag-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_I64LE "tag_datatype";
         H5T_STD_U64LE "tag_handle";
         H5T_STRING {
            STRSIZE 256;
            STRPAD H5T_STR_NULLTERM;
            CSET H5T_CSET_ASCII;
            CTYPE H5T_C_S1;
         } "name";
      }
      DATASPACE  SIMPLE { ( 1 ) / ( H5S_UNLIMITED ) }
      DATA {
      (0): {
            4,
            8589939592,
            "some_array"
         }
      }
   }
   DATASET "damsel-vector-handle-containers" {
      DATATYPE  H5T_COMPOUND {
         H5T_STD_U64LE "handle";
         H5T_STD_I32LE "quantity";
         H5T_STD_I32LE "blocknumber";
         
         H5T_OPAQUE;
         OPAQUE_TAG "";
       "content";
      }
      DATASPACE  SIMPLE { ( 0 ) / ( 0 ) }
      DATA {
      }
   }
   DATASET "some_array-200001392" {
      DATATYPE  H5T_IEEE_F32LE
      DATASPACE  SIMPLE { ( 8 ) / ( 8 ) }
      DATA {
      (0): 0, 0, 0, 0, 0, 0, 0, 0
      }
   }
   DATASET "some_array-300001391" {
      DATATYPE  H5T_IEEE_F32LE
      DATASPACE  SIMPLE { ( 7 ) / ( 7 ) }
      DATA {
      (0): 1, 1, 1, 1, 1, 1, 1
      }
   }
}
}
*/
