/* This example mimic PnetCDF APIs to read two 1D array of float and int types.
 * The PnetCDF objects created in this example are:
 * 1. One file (corresponding to Damsel model)
 * 2. One dimension (corresponding to a Damsel container)
 * 3. Two 1D array variables using the dimension from 2
 *    3.1 flt_var1 (corresponding to a Damsel tag of type float)
 *    3.2 int_var1 (corresponding to a Damsel tag of type int)
 * 4. Open the file written in two_array1d.c
 * 5. Query for the tags for flt_var1 and int_var1
 * 6. Fill in the arrays with data read  
 * 5. close the file (corresponding to close containers and model)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <mpi.h>
#include <damsel.h>
#include "damsel-internal.h"

/* define macros used in this program */
#define DIM_LEN 15

#define WRITE_MODE    0
#define READ_MODE     1
#define NUM_PER_CHUNK 32

/* C structures used in this program */
typedef struct {
    char              name[128];
    damsel_container  container;
    int               num_tags;
    damsel_handle    *tags;
} container_list;

typedef struct {
    MPI_Comm        comm;
    damsel_model    model;
    int             num_containers;
    container_list *containers;
} file_descriptor;

typedef file_descriptor * file_dscpt;

/* function declaration */
static damsel_err_t mimic_ncmpi_file_open(MPI_Comm comm, char *filename, int cmode, file_dscpt *fd);
//static damsel_err_t mimic_ncmpi_def_dim(file_dscpt fd, char *name, int dim_len, damsel_container *dim_id);
//static damsel_err_t mimic_ncmpi_def_var(file_dscpt fd, char *name, damsel_data_type type, damsel_container dim_id, void *buf, damsel_handle *var_id);
//static damsel_err_t mimic_ncmpi_enddef(file_dscpt fd);
static damsel_err_t mimic_ncmpi_iget(file_dscpt fd, damsel_request_t *req);
static damsel_err_t mimic_ncmpi_wait(file_dscpt fd, damsel_request_t req, damsel_status_t *status);
static damsel_err_t mimic_ncmpi_close(file_dscpt *fd);

#define CHECK_ERR(e) {if ((e.id!=DMSL_OK.id)) { printf("Error at line=%d msg=%s\n", __LINE__,DMSLerror_str(e)); goto fn_exit;}}

/*----< main() >--------------------------------------------------------------*/
int main(int argc, char **argv) {
    int i, rank, nprocs, *int_buf;
    float *flt_buf;
    damsel_err_t err;

#ifndef DAMSEL_SINGLETHREAD
    int res;
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
    assert(res == MPI_THREAD_MULTIPLE);
#else
    MPI_Init(&argc, &argv);
#endif

    /* initialize Damsel environment */
    damsel_library lib;
    err = DMSLlib_init(&lib);
    CHECK_ERR(err)

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (argc != 2) {
        if (!rank) printf("Usage: %s filename\n",argv[0]);
        goto fn_exit;
    }

    /* create a new file, here we will open a file for read */
    file_dscpt fd;
    err = mimic_ncmpi_file_open(MPI_COMM_WORLD, argv[1], READ_MODE, &fd);
    CHECK_ERR(err)
    damsel_model_internal *ms = (damsel_model_internal *)(fd->model);
    assert(ms != NULL);

    //DMSLmodel_populate(fd->model); 
   
    /************************************************************/
    /*Now query/inquire the file for all the tags i.e. flt_var1*/
    damsel_handle fd_flt_var1_tag;
    err = DMSLselect_tag_by_name(fd->model, "flt_var1", &fd_flt_var1_tag);
    damsel_container fd_flt_var1_tag_container;
    err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, fd_flt_var1_tag, 1, 1, &fd_flt_var1_tag_container);
    
    /*Now query/inquire the file for all the tags i.e. int_var1*/
    damsel_handle fd_int_var1_tag;
    err = DMSLselect_tag_by_name(fd->model, "int_var1", &fd_int_var1_tag);
    damsel_container fd_int_var1_tag_container;
    err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, fd_int_var1_tag, 1, 1, &fd_int_var1_tag_container);
 
    /*We can use either of the tag handles to get the sequence of handles in a container, as both tags were mapped to the same set of handles inside a container*/
    damsel_container fd_container = DMSLselect_handles_with_values(fd->model, fd_flt_var1_tag);
    int count = DMSLcontainer_count(fd_container); 
    
    /************************************************************/
    /*Now create in-memory handles, for container of damsel_handles use count */
    damsel_handle mem_flt_var1_tag = 1000;
    damsel_container mem_flt_var1_tag_container;
    err = DMSLcontainer_create_sequence(fd->model, mem_flt_var1_tag, 1, 1, &mem_flt_var1_tag_container);

    damsel_handle mem_int_var1_tag = 2000;
    damsel_container mem_int_var1_tag_container;
    err = DMSLcontainer_create_sequence(fd->model, mem_int_var1_tag, 1, 1, &mem_int_var1_tag_container);
/*
    damsel_container mem_container;
    err = DMSLcontainer_create_sequence(fd->model, 10000, count, 1, &mem_container);
  */  
    /************************************************************/

    /*map memory side and file side handles for the tag, but these have to be in a container*/
    DMSLmodel_map_handles(mem_flt_var1_tag_container, fd_flt_var1_tag_container);
    DMSLmodel_map_handles(mem_int_var1_tag_container, fd_int_var1_tag_container);

    //DMSLmodel_map_handles(mem_container, fd_container);
    int cpos;
    for(cpos = 0; cpos<count; cpos++) {
        //printf("fd_handle %d = %lld, mem_handle = %lld\n", cpos, DMSLcontainer_handle_at_position(fd_container, cpos), DMSLcontainer_handle_at_position(mem_container, cpos));
        //printf("\n");
    }
/* 
    DMSLmodel_map_tag(flt_buf, mem_container, &mem_flt_var1_tag);
    DMSLmodel_map_tag(int_buf, mem_container, &mem_int_var1_tag);
   
    DMSLmodel_map_handles(mem_container, fd_container);
  */  //DMSLcontainer_release(fd_container);
   
    /* initialize local array of size local_len */
    int local_len = count / nprocs;
    if (rank < count % nprocs) local_len++;
    flt_buf = (float*) malloc(local_len * sizeof(float));
    int_buf = (int*)   malloc(local_len * sizeof(int));
    
    damsel_container mem_container;
    err = DMSLcontainer_create_sequence(fd->model, 10000, local_len, 1, &mem_container);
    
    damsel_handle *read_handles;
    read_handles = (damsel_handle*)   malloc(local_len * sizeof(damsel_handle));

    int c, gh=0;
    for(c = rank; c<count; c+=nprocs) {
        printf("handle %d = %llx\n", c, DMSLcontainer_handle_at_position(fd_container, c));
        read_handles[gh] = DMSLcontainer_handle_at_position(fd_container, c);
        gh++;
}

    damsel_container read_container;
    err = DMSLcontainer_create_vector((damsel_model)ms->attached_file, read_handles, local_len, &read_container);

//    damsel_container read_container;
  //  err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, DMSLcontainer_handle_at_position(fd_container, rank), local_len, nprocs, &read_container);
    CHECK_ERR(err)
    DMSLmodel_map_handles(mem_container, read_container);
    err = DMSLmodel_map_tag(flt_buf, mem_container, &mem_flt_var1_tag);
    CHECK_ERR(err)
    err = DMSLmodel_map_tag(int_buf, mem_container, &mem_int_var1_tag);
    CHECK_ERR(err)
    
    /************************************************************/
    
    /* Now we can tell Damsel to start writing the data */
    damsel_request_t req;
    err = mimic_ncmpi_iget(fd, &req);
    CHECK_ERR(err)

    /* Other computations can be performed here, i.e. between the post of
     * asynchronous transfer and wait, as long as the registered data buffer,
     * buf, is not touched */

    damsel_status_t status;
    err = mimic_ncmpi_wait(fd, req, &status);
    CHECK_ERR(err)

    err = mimic_ncmpi_close(&fd);
    //printf("");
    for (i=0; i<local_len; i++) {
        printf("f = %f, i = %d %lld %lld\n", flt_buf[i], int_buf[i], DMSLcontainer_handle_at_position(fd_container, i), DMSLcontainer_handle_at_position(mem_container, i));
        //printf("%d\n", int_buf[i]);
     }
    free(flt_buf);
    free(int_buf);
    free(read_handles);
fn_exit:
    /* exit Damsel environment */
    if (lib != DAMSEL_LIBRARY_INVALID) {
        err = DMSLlib_finalize(lib);
        CHECK_ERR(err)
    }

    MPI_Finalize();
    return 0;
}

/*----< mimic_ncmpi_file_open() >-------------------------------------------*/
/* 1. create a file descriptor, to store the following Damsel objects.
 * 2. make a copy of MPI communicator
 * 3. create a Damsel model handle
 * 4. set the access mode as READ for the model
 */
static
damsel_err_t mimic_ncmpi_file_open(MPI_Comm    comm,
                                     char       *filename,
                                     int         cmode, /* create mode */
                                     file_dscpt *fd)    /* OUT */
{
    damsel_err_t err;

    *fd = (file_descriptor*) malloc(sizeof(file_descriptor));
    (*fd)->num_containers = 0;
    (*fd)->containers     = NULL;

    /* make a copy of comm in fd */
    MPI_Comm_dup(comm, &(*fd)->comm);

    /* create a model, a in-memory store */
    err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &(*fd)->model);
    if (err.id != DMSL_OK.id) return err;
    /* make sure the file does not exist, ignore error if any */

    /* set the file access mode */
    damsel_trait_bundle tb;
    if (err.id != DMSL_OK.id) return err;
    err = DMSLtrait_create(&tb);
    if (err.id != DMSL_OK.id) return err;

    if (cmode != READ_MODE) return DMSL_ECMODE;

    err = DMSLtrait_put(tb, "mode", "READ");
    if (err.id != DMSL_OK.id) return err;
    /* register the filename and access mode to model (in-memory store) */
    err = DMSLmodel_attach((*fd)->model, filename, comm, tb);
    if (err.id != DMSL_OK.id) return err;

    err = DMSLtrait_release(tb);
    if (err.id != DMSL_OK.id) return err;
    DMSLmodel_populate((*fd)->model);
    
    return DMSL_OK;
}

/*----< mimic_ncmpi_iget() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_iget(file_dscpt        fd,
                              damsel_request_t *req)
{
    damsel_err_t err;
    err = DMSLmodel_transfer_async(fd->model, DAMSEL_TRANSFER_TYPE_READ, req);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_wait() >--------------------------------------------------*/
static
damsel_err_t mimic_ncmpi_wait(file_dscpt        fd,
                              damsel_request_t  req,
                              damsel_status_t  *status)
{
    damsel_err_t err;
    /* wait for the I/O transfer to complete */
    err = DMSLmodel_wait(req, status);
    if (err.id != DMSL_OK.id) return err;

    return DMSL_OK;
}

/*----< mimic_ncmpi_close() >-------------------------------------------------*/
/* Since the file descriptor stores all Damsel objects, we can retrieve them
 * and release the resource used.
 */
static
damsel_err_t mimic_ncmpi_close(file_dscpt *fd)
{
    int i;
    damsel_err_t err;

    /* close all container handles */
    for (i=0; i<(*fd)->num_containers; i++) {
        err = DMSLcontainer_close((*fd)->containers[i].container);
        if (err.id != DMSL_OK.id) return err;
    }
    /* close in-memory store */
    err = DMSLmodel_close((*fd)->model);
    if (err.id != DMSL_OK.id) return err;

    free(*fd);

    return DMSL_OK;
}

/*
Compile:
mpicc -c -g -I/path/to/Damsel/include two_array1d_read.c

Link:
mpicc -o two_array1d two_array1d_read.o -L/path/to/Damsel/lib -L/path/to/HDF5/lib -L/path/to/PnetCDF/lib -ldamsel -lhdf5 -lhdf5_hl -lm -lz -lpnetcdf

Run: (on 2 process in parallel)
mpiexec -n 2 two_array1d_read output/dmsl.h5
*/
