
#include <assert.h>
#include <string.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-debug.h"
#include "damsel-ndc.h"

/* used for qsort and bsearch for sorting and searching handles */
int compare_handles(const void *a, const void *b)
{
  return ( ((damsel_handle_index*)a)->handle - ((damsel_handle_index*)b)->handle );
}

int get_offset_in_container(damsel_container_internal *, damsel_handle);

int handle_containers_overlap(damsel_container l_id, damsel_container r_id) {
  push_ndc("handle_containers_overlap");
  assert_valid_handle_container_id(l_id);
  assert_valid_handle_container_id(r_id);
  /* the quick path: are they the same container? */
  if(l_id == r_id) {
    printf_ndc("handle containers are equal, so they overlap. left container = right container = %p", l_id);
    pop_ndc();
    return 1;
  }

  /* now the slow path. this is going to be really slow for large
     containers!
  */
  // iterate over every element e of l.
  // does r containt e? if so, then return true.

  damsel_container_internal *lhc = (damsel_container_internal *)l_id;
  assert(lhc != NULL);
  int li;
  for(li=0; li<lhc->quantity; li++) {
    damsel_handle lh = DMSLcontainer_handle_at_position(l_id, li);
    damsel_container_internal *rhc = (damsel_container_internal *)r_id;
    assert(rhc != NULL);
    int ri;
    for(ri=0; ri < rhc->quantity ; ri++) {
      damsel_handle rh = DMSLcontainer_handle_at_position(r_id, ri);
      if(lh == rh) {
        printf_ndc("found overlap of handle container id %p with container id %p", l_id, r_id);
        printf_ndc("overlapping on position %d in lefthand container, %d in righthand container", li, ri);
        printf_ndc("handle at these positions is %llx", lh);
        pop_ndc();
        return 1;
      }
    }
  }
  printf_ndc("handle containers do not overlap at all.");
  pop_ndc();
  return 0;
}
/* this will allocate memory associated with the store. I need to free it when
   I free the store. Maybe (for long lived memory stores with changing
   handle containers?) I will need separate earlier deallocation too.

   This implementation is for in-memory use. What does a file based
   sequence look like?

*/

damsel_err_t DMSLcontainer_create_sequence(damsel_model model,
                                           damsel_handle base,
                                           unsigned quantity,
                                           size_t step,
                                           damsel_container *container) {
  push_printf_ndc("DMSLcontainer_create_sequence, model=%p", model);
  if (model == NULL) return DMSL_EMODEL;

  assert_valid_store_id(model); /* return DMSL_EMODEL; */

  printf_ndc("registering sequence of handles");
  printf_ndc("base: %llx (hex)", base);
  printf_ndc("step: %zd (decimal)", step);
  printf_ndc("quantity: %d (decimal)", quantity);

  // now we need to store this information.
  // lets have a linked list of known handles in the store.
  // later we can annotate that list.
  damsel_seq_handle_container *hseq = malloc(sizeof(damsel_seq_handle_container));
  if (hseq == NULL) return DMSL_EMALLOC;

  hseq->container.type = DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE;
  hseq->container.model = model;
  model_take((damsel_store_internal *)model);
  hseq->container.quantity = quantity;
  hseq->container.refcount = 1; // the reference that will be returned to the application
  hseq->container.written = DAMSEL_UNWRITTEN;
  hseq->base = base;
  hseq->step = step;

  damsel_store_internal *s = (damsel_store_internal *)model;
  printf_ndc("registering into store %p", s);
  assert(s->type == DAMSEL_STORETYPE_MEMORY || s->type == DAMSEL_STORETYPE_HDF5);

//  linkedlist_prepend(&s->first_container, hseq);

  printf_ndc("created sequence container with id %p, in model %p", hseq, (void *)hseq->container.model);
  pop_ndc();
  *container = (damsel_container)hseq;
  return DMSL_OK;
}


/* when we write this out to an HDF5 file, store this as a variable
   length sequence. (TODO) for now vector containers are stored in fixed
   length arrays of VECTOR_MAX elements, so cannot exceed that number of entries
   if the container is going to a file. this limit does not apply if this
   container does not end up serialised in the file.

   question: should we be copying the contents here or should the application
   leave that memory allocated until after some later point?

   tim says not-copy.
*/

/* 'contents' must not be stack-allocated.  it must be memory passed to
 * DMSLcontainer_create_vector: create_vector will manage the memory */
damsel_err_t
DMSLcontainer_create_vector(damsel_model model,
                            // damsel_handle collh
                            damsel_handle *contents,
                            unsigned quantity,
                            damsel_container *container) {
  push_printf_ndc("DMSLcontainer_create_vector, model=%p", model);
  assert(model); /* if (model == NULL) return DMSL_EMODEL; */
  assert_valid_store_id(model);
  damsel_vector_handle_container *hvec = malloc(sizeof(damsel_vector_handle_container));
  assert(hvec); /* if (hvec == NULL) return DMSL_EMALLOC; */

  hvec->container.type = DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR;
  hvec->container.quantity = quantity;
  hvec->container.model = model;
  model_take((damsel_store_internal *)model);
  hvec->container.refcount = 1; // the reference that will be returned to the application
  hvec->container.written = DAMSEL_UNWRITTEN;

  hvec->contents = calloc(quantity, sizeof(damsel_handle));
  memcpy(hvec->contents, contents, quantity*sizeof(damsel_handle));

#ifdef USE_BSEARCH 
  /* needed for bsearch at get_offset_in_container_vector() */
  {
    int i;
	hvec->contents_sorted = (damsel_handle_index *)malloc(quantity*sizeof(damsel_handle_index));
	for (i=0; i<quantity; i++) {
	    hvec->contents_sorted[i].handle = contents[i];
		hvec->contents_sorted[i].index = i;
	}
	qsort (hvec->contents_sorted, quantity, sizeof(damsel_handle_index), compare_handles);
  }
#endif
  
  damsel_store_internal *s = (damsel_store_internal *)model;

  assert(s->type == DAMSEL_STORETYPE_MEMORY || s->type == DAMSEL_STORETYPE_HDF5);

//  linkedlist_prepend(&s->first_container, hvec);

  printf_ndc("created vector container with id %p, model=%p, quantity=%u", hvec, (void *)hvec->container.model, quantity);

  pop_ndc();
  *container = (damsel_container)hvec;

  return DMSL_OK;
}

damsel_container DMSLcontainer_tree_create(damsel_model model_id,
                            damsel_collection_type coll_type
                           ) {
  push_ndc("in DMSLcontainer_tree_create");
  assert(model_id != NULL);
  assert(coll_type == DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR
       ||coll_type == DAMSEL_HANDLE_COLLECTION_TYPE_SET);

  damsel_tree_handle_container *hcoll = malloc(sizeof(damsel_tree_handle_container));
  assert(hcoll != NULL);

  hcoll->container.type = DAMSEL_HANDLE_CONTAINER_TYPE_TREE;
  hcoll->container.quantity = 0; // this will increase as we add using coll_add
  hcoll->container.model = model_id;
  model_take((damsel_store_internal *)model_id);
  hcoll->container.refcount = 1; // the reference that will be returned to the application
  hcoll->container.written = DAMSEL_UNWRITTEN;
  hcoll->first_hc = NULL;

//  linkedlist_prepend(&s->first_container, hcoll);

  printf_ndc("tree container created with id %p in model %p", hcoll, model_id);
  pop_ndc();
  return (damsel_container)hcoll;
}

damsel_err_t DMSLcontainer_tree_add(damsel_container coll_hc_id,
                         damsel_container contained_hc_id) {
  push_printf_ndc("DMSLcontainer_tree_add, tree container = %p, content container = %p", coll_hc_id, contained_hc_id);
  damsel_tree_handle_container *hc = (damsel_tree_handle_container *)coll_hc_id;
  assert(hc->container.type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE);
  linkedlist_prepend(&hc->first_hc, contained_hc_id);
  hc->container.quantity += ((damsel_container_internal *)contained_hc_id)->quantity;
  DMSLcontainer_take(contained_hc_id);
  pop_ndc();
  return DMSL_OK;
}

damsel_err_t DMSLcontainer_tree_add_fast(damsel_container container_id, damsel_handle_ptr handles, size_t count) {
  damsel_err_t err;
  damsel_model model = ( (damsel_container_internal *) container_id)->model;
  damsel_container hc_id;
  err = DMSLcontainer_create_vector(model, handles, count, &hc_id);
  assert_valid_handle_container_id(hc_id);
  damsel_err_t rc = DMSLcontainer_tree_add(container_id, hc_id);
  DMSLcontainer_release(hc_id);
  return rc;
}

damsel_err_t DMSLcontainer_take(damsel_container c) {
  printf_ndc("DMSLcontainer_take %p", c);

  if (!c)
      return DMSL_ERROR_GENERAL;
	  
  damsel_container_internal *ci = (damsel_container_internal *)c;
  ci->refcount++;
  
  return DMSL_OK;
}

static void release_tree_container_node(void *m) {
  DMSLcontainer_release(m);
}

static void release_tree_container(damsel_container_internal *ci) {
  assert(ci!=NULL);
  assert(ci->type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE);
  damsel_tree_handle_container *tci = (damsel_tree_handle_container *)ci;
  linkedlist_free(tci->first_hc, release_tree_container_node);
}

damsel_err_t DMSLcontainer_release(damsel_container c) {
  void *ctx = push_printf_ndc("DMSLcontainer_release %p", c);
  damsel_container_internal *ci = (damsel_container_internal *)c;
  assert(ci->refcount > 0);
  ci->refcount--;
  printf_ndc("refcount is now %d", ci->refcount);
  if(ci->refcount == 0) {
    printf_ndc("container refcount is 0");
    if(ci->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR) {
      damsel_vector_handle_container *vc = (damsel_vector_handle_container *)ci;
      free(vc->contents);
#ifdef USE_BSEARCH 
      free(vc->contents_sorted);
#endif	  
    }

    if(ci->type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE) {
      release_tree_container(ci);
    }

    model_release((damsel_store_internal *)ci->model);
    free(ci);
  }

  pop_ndc_context(ctx);
  return DMSL_OK;
}

damsel_err_t DMSLcontainer_close(damsel_container c) {
  return DMSLcontainer_release(c);
}

void set_written_container(damsel_container c) {
  damsel_container_internal *ci = (damsel_container_internal *)c;
  ci->written = DAMSEL_WRITTEN;
}

damsel_err_t DMSLcontainer_sequence_get_contents(damsel_model model_id,
                                                 damsel_container c,
                                                 damsel_handle_ptr handle_ptr,
                                                 size_t *count,
                                                 size_t *incr) 
{
  damsel_container_internal *hc = (damsel_container_internal *)c;
//  push_printf_ndc("DMSLcontainer_sequence_get_contents, container id %p", hc);
  if(hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    damsel_seq_handle_container *hsc = (damsel_seq_handle_container *)hc;
//    pop_ndc();
    *handle_ptr = hsc->base;
    *count = hc->quantity;
    *incr = hsc->step;
    return DMSL_OK;
  }
  else {
//    pop_ndc();
//    push_printf_ndc("DMSLcontainer_sequence_get_contents, container id %p: wrong container type.", hc);
//    pop_ndc();
    assert(0);
    damsel_err_t err;
    err.id = 1;
    return err;
  }
}

damsel_err_t DMSLcontainer_vector_get_contents(damsel_model model_id,
                                               damsel_container c,
                                               damsel_handle_ptr * const handle_ptr_ptr,
                                               size_t *count) 
{
  damsel_container_internal *hc = (damsel_container_internal *)c;
//  push_printf_ndc("DMSLcontainer_vector_get_contents, container id %p", hc);
  if (hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR) {
    damsel_vector_handle_container *hvc =
      (damsel_vector_handle_container *)hc;
//    pop_ndc();
    *handle_ptr_ptr = hvc->contents;
    *count = hc->quantity;
    return DMSL_OK;
  }
  else {
//    pop_ndc();
//    push_printf_ndc("DMSLcontainer_vector_get_contents, container id %p: wrong container type.", hc);
//    pop_ndc();
    assert(0);
    damsel_err_t err;
    err.id = 1;
    return err;
  }
}

/*! \brief get contents of tree container 
 * If NULL is passed in for current container, returns first container in tree; for subsequent
 * calls, returns container after currently pointed-to container; returns NULL if there are no
 * more containers in tree, or (if NULL is passed in) if tree is empty.
 */
damsel_err_t DMSLcontainer_tree_get_contents(damsel_model model_id,
                                             damsel_container c,
                                             void *hc_node,
                                             damsel_container *hc_ptr) 
{
  damsel_container_internal *hc = (damsel_container_internal *)c;
//  push_printf_ndc("DMSLcontainer_tree_get_contents, container id %p", hc);
  if(hc->type != DAMSEL_HANDLE_CONTAINER_TYPE_TREE) {
//    pop_ndc();
    push_printf_ndc("DMSLcontainer_tree_get_contents, container id %p: wrong container type.", hc);
//    pop_ndc();
    assert(0);
    damsel_err_t err;
    err.id = 1;
    return err;
  }
  else {
    damsel_tree_handle_container *thc = (damsel_tree_handle_container *)hc;
    linkedlist **llptr = ((linkedlist**)hc_node);
    if (!*llptr) {
      *llptr = thc->first_hc;
      *hc_ptr = (*llptr)->head;
      return DMSL_OK;
    }
    else {
      *llptr = (*llptr)->next;
      if (*llptr) *hc_ptr = (*llptr)->head;
      else *hc_ptr = NULL;
      return DMSL_OK;
    }
  }
}

int get_offset_in_container_sequence(damsel_container_internal *hc, damsel_handle src_handle) {
  printf_ndc("found correct handle container, type=sequence");
  damsel_seq_handle_container *hsc = (damsel_seq_handle_container *)hc;
  // then this is the source container that the mapping refers to
  // so check if the container contains src_handle

  if (hc->quantity <= 0) return -1;
  damsel_handle off = src_handle - hsc->base;
  if (off < 0 || off % hsc->step) return -1;

  off /= hsc->step;
  if (off >= hc->quantity) return -1;

  printf_ndc("returning offset %d", (int)off);
  return off;
}

int get_offset_in_container_vector(damsel_container_internal *hc, damsel_handle src_handle) {
  printf_ndc("found correct handle container, type=vector");
  damsel_vector_handle_container *hvc = (damsel_vector_handle_container *)hc;
  
#ifdef USE_BSEARCH 
  /* use bsearch for better performance */
  damsel_handle_index target, *found=NULL;
  target.handle = src_handle;
  found = (damsel_handle_index *)bsearch(&target, hvc->contents_sorted, 
          hc->quantity, sizeof(damsel_handle_index), compare_handles);
  if (found) return found->index;
#else   
  int i;
  for(i=0; i<hc->quantity; i++) {
    if(hvc->contents[i] == src_handle) {
      printf_ndc("returning offset %d", i);
      return i;
    }
  }
#endif
  
  return -1;
}

int get_offset_in_container_tree(damsel_container_internal *hc, damsel_handle src_handle) {
  printf_ndc("found correct handle container, type=coll");
  damsel_tree_handle_container *hcc = (damsel_tree_handle_container *)hc;
  // now step through each member container, offsetting as appropriate
  int offset=0;
  linkedlist *here = hcc->first_hc;
  while(here != NULL) {
    damsel_container_internal *here_hc;
    here_hc = (damsel_container_internal *)here->head;
    int i = get_offset_in_container(here_hc, src_handle);
    if(i>=0) {
      int rv=i+offset;
      printf_ndc("returning offset %d",rv);
      return rv;
    }
    offset+=here_hc->quantity;
    here = here->next;
  }
  return -1;
}

int get_offset_in_container(damsel_container_internal *hc, damsel_handle src_handle)
{
  void *ctx = push_ndc("get_offset_in_container");
  if(hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    int o = get_offset_in_container_sequence(hc, src_handle);
    pop_ndc_context(ctx);
    return o;
  } else if ( hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR ) {
    int o = get_offset_in_container_vector(hc, src_handle);
    pop_ndc_context(ctx);
    return o;
  } else if ( hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_TREE ) {
    int o = get_offset_in_container_tree(hc, src_handle);
    pop_ndc_context(ctx);
    return o;
  } else {
    assert(1==0 && "unknown handle container type");
  }
  printf_ndc("no mapping found for handle %llx", src_handle);
  pop_ndc_context(ctx);
  return -1;
}

int find_handle_in_container(damsel_container_internal *hc,
  damsel_handle h) {
  assert_valid_handle_container_id(hc);
  if(hc->type == DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE) {
    damsel_seq_handle_container *seq = (damsel_seq_handle_container *)hc;
    assert(seq->step > 0);
    damsel_handle nh = h - seq->base;  // normalise to start at 0
    if((nh % seq->step) != 0) return -1; // offset is not divisible by step.
    int offset = nh / seq->step;
    if(offset < 0 || offset >= hc->quantity) return -1;
    return offset;
  } else { // slow path that will work for any representation
    int ix;
    for(ix=0; ix<hc->quantity; ix++) {
      damsel_handle candidate = DMSLcontainer_handle_at_position((damsel_container)hc, ix);
      if(h == candidate) { assert(ix >= 0); return ix; }
    }
    return -1;
 }
}

