/* this is for implementation-internal definitions that are not exposed
through the public API */

#ifndef __DAMSEL_INTERNAL_H
#define __DAMSEL_INTERNAL_H

/* switch between USE_BSEARCH and NOT_USE_BSEARCH, for debug purpose only */
#define USE_BSEARCH

#ifdef __cplusplus
extern "C" {
#endif

/* the largest number of handles that can be stored in a vector container in
   a file. this value is part of the definition of the present file format -
   changing to a different value will render files incompatible */
#define VECTOR_MAX 128

/* Compile-time "assert" macro */
#define compile_assert(e)     do { enum { compile_assert__ = 1 / (e) }; } while(0)

/* names of datasets used to store metadata */
#define DAMSEL_DATASET_CONNECTIVITY "damsel-entity-connectivity"
#define DAMSEL_DATASET_CONTAINER_SEQ "damsel-seq-handle-containers"
#define DAMSEL_DATASET_CONTAINER_VECTOR "damsel-vector-handle-containers"
#define DAMSEL_DATASET_TAG "damsel-tag-containers"
#define DAMSEL_DATASET_COLLECTION "damsel-collections"
#define DAMSEL_DATASET_HANDLE_ALLOCATOR "damsel-handle-allocator"
#define DAMSEL_DATASET_VERSIONING "damsel-versions"
#define DAMSEL_DATASET_VERSIONING_DEPENDENCY "damsel-versions-dependencies"
#define DAMSEL_DATASET_NUM_SUBFILES "damsel-num-subfiles"
#define DAMSEL_DATASET_SUBFILE_INDEX "damsel-subfile-index"


/* linked list for use in various places */
typedef struct linkedlist {
  void *head;
  struct linkedlist *next;
} linkedlist;

/*! this is an in-memory representation of a handle container. the type
    defines which structure defines the rest of the content. */

typedef struct {

  /*! the model that this container belongs to. */
  damsel_model model;

  /*! type values come from DAMSEL_HANDLE_CONTAINER_TYPE_* defines. */
  int type;

  unsigned quantity;

  /*! reference count */
  int refcount;

  /*! has this container already be written out to disk? */
  int written;

} damsel_container_internal;

/*! this is an in-memory representation of a handle container that is
    defined as a sequence. */
typedef struct damsel_seq_handle_container {
  damsel_container_internal container;
  damsel_handle base;
  size_t step;
} damsel_seq_handle_container;

typedef struct {
  damsel_handle handle;
  int index;
} damsel_handle_index;

/*! this is an in-memory representation of a handle container that is
    defined as a vector of explicit handles. */
typedef struct damsel_vector_handle_container {
  damsel_container_internal container;
  damsel_handle *contents;
  damsel_handle_index *contents_sorted;
} damsel_vector_handle_container;

typedef struct damsel_tree_handle_container {
  damsel_container_internal container;
  linkedlist /* of handle container IDs */ *first_hc;
} damsel_tree_handle_container;

/* write status used to indicate if an in-memory structure has
   been written out to the file already (perhaps in a previous
   session */
#define DAMSEL_UNWRITTEN 0x787
#define DAMSEL_WRITTEN 0x339

#define DAMSEL_TAG_MAGIC 0x121

typedef struct {
  int magic;
  damsel_handle handle;
  char *name;
  damsel_data_type type;
  int written; 
} damsel_tag_internal;

typedef struct damsel_collection_internal {
  int type;
  damsel_handle collection_handle;
  damsel_container container;
} damsel_collection_internal;


/*! this structure represents a queued write operation from the *data
   memory buffer to the tag specified, at values specified by the
   'handles' handle container */
typedef struct damsel_cached_data_internal {
  damsel_container handles_id;

  /*! this tag handle is in the handle space of the source store, not of
      any destination store that it might me mapped to. */
  damsel_handle tag_handle;

  void * data;

  struct damsel_cached_data_internal *next;
} damsel_cached_data_internal;


/*! this is a structure to allow allocation of in-file handles later on
    according to the file storage layer, rather than requiring the application
    to do so. the supplied handle should be a handle to a handle container -
    i.e. something that can be passed as the source handle argument to
    DMSLmap.

    what about handles contained *within* this handle container? should I
    be checking those have been mapped or not? need to think about that
    some more.
*/
typedef struct damsel_metadata_handle {
  struct damsel_metadata_handle *next;
  void (*permutation_callback)(damsel_container, damsel_container);
  damsel_container container_id;
} damsel_metadata_handle;


typedef struct {
  /* here I'm storing the list of registered handles in an unordered
     linked list (so inefficient lookups...)
     later on it might be useful to store it in some other form? */

  /*! takes values from DAMSEL_STORETYPE_* constants */
  int type;

  // linkedlist *first_container; /* of damsel_container_internal */
  int refcount;
} damsel_store_internal;

#define DAMSEL_STORE_MAPPING_MAGIC 14372
typedef struct {
  int magic;
  // damsel_store *src_store; - no more src store because both source and
  // destination stores are now implied by the containing model; and are
  // also (redundantly) stored in the containers.

  damsel_container app_container_id;

  damsel_container file_container_id;

} damsel_store_mapping;

typedef struct {
  damsel_store_internal store;
  damsel_cached_data_internal *first_cached_data;
  struct damsel_attached_file *attached_file;
#ifdef DAMSEL_SUBFILE
  struct damsel_attached_file *attached_subfile;
  int num_subfiles;
#endif
  linkedlist /* of damsel_internal_entity_definition */ *first_entity_definition;

  /*! a list of handles that the application will not have mapped but
      should be put in the file anyway - so they will need mapping to some
      handles of the file but the application doesn't care what actual
      value because its not directly exposed... */
  damsel_metadata_handle *first_metadata_handle;
  linkedlist *mappings; /* of damsel_store_mapping */

  /*! global mappings - both for this rank, and during transfer, for
      all ranks. */
  linkedlist *global_mappings;

  /*! used to track phases in writing data where first new
      mappings may be established, and secondly where they
      may not, as so may be safely exchanged between ranks.

      benc is unsure that this state is needed in this
      implementation.
  */
  int more_mappings_allowed;

  linkedlist *first_tag_definition; /* of damsel_tag_internal */

  damsel_transfer_type direction; /* for assertions */

  linkedlist /*of damsel_cached_data_internal */ *transaction_stack;

  linkedlist /*of damsel_internal_chunk */ *first_chunk;

  linkedlist /*of damsel_collection_internal */ *first_collection;

} damsel_model_internal;

/* A chunk is a block on disk into which data will be written.
   At the moment, a chunk stores data from exactly one tag, on
   exactly one file-side container. */

#define DAMSEL_CHUNK_MAGIC 0x82811
typedef struct {
  int magic;
  damsel_tag_internal *tag;
  damsel_container_internal *container;
} damsel_internal_chunk;

#define DAMSEL_GLOBAL_MAPPING_MAGIC 0x62382

typedef struct {
  int magic;
  int rank;
  damsel_container app_container;
  int *proc_ranks;
  damsel_handle_ptr remote_handles;

  /*! if true, then when this mapping is freed, then
      the memory for proc_ranks and remote_handles
      will be freed too */
  int manage_memory_internally;

} damsel_internal_global_mapping;

typedef enum {
  DAMSEL_HDF_DATATYPE_ENUM_SHORTINT,
  DAMSEL_HDF_DATATYPE_ENUM_DOUBLE
} damsel_hdf_datatype_enum;

// call at initialisation to initialise the type system.
void init_types();

int inner_assert_valid_tag_handle(damsel_handle h, damsel_model_internal *store);

char *tagname(damsel_model_internal *store, damsel_handle tag_handle);

long damsel_type_to_serializable_type(damsel_data_type damsel_type);
damsel_data_type damsel_serializable_type_to_damsel_type(long it);

/* this ties in types between the damsel-types code and at least the HDF5
   specific code. */

extern linkedlist *defined_types;

/* is the library initialised or not? 0 for not. probably should become
   a counter eventually... */
extern int globals_initialized;

damsel_err_t handle_map(damsel_model_internal *model, damsel_transfer_type direction,
  damsel_handle src_handle, damsel_handle *dest_handle);

/*! returns the index of the specified handle in the specified handle
    container, or -1 if that handle is not contained in the container. */
int find_handle_in_container(damsel_container_internal *hc,
  damsel_handle h);

int test_no_container_with_this_handle(damsel_model model, damsel_handle h);

/*! checks if two handle containers in the same store overlap - that is,
    if there is some handle h that is contained in both containers. returns
    1 if there is overlap, 0 otherwise. */
int handle_containers_overlap(damsel_container l_id, damsel_container r_id);

void store_entities(damsel_model_internal *s, struct damsel_attached_file *fs);

int linkedlist_length(linkedlist *l);
void linkedlist_prepend(linkedlist **la, void *entry);
void linkedlist_append(linkedlist **la, void *entry);
void linkedlist_free(linkedlist *ll, void (*free_entry)(void *m) );
linkedlist *linkedlist_filter(linkedlist *ll, int (*filter)(void *e));



void entities_initalize_for_read(struct damsel_attached_file *s);

void internal_record_pushed_container(struct damsel_attached_file *f,
  damsel_container app_container,
  damsel_container file_container,
  damsel_handle file_handle,
  int remote);

/*! set handleptr to NULL to push only and return no handle. */
damsel_err_t internal_push_container(damsel_container app_container_id, damsel_handle *handleptr);

damsel_handle internal_push_container_lookup(damsel_container app_container_id);
damsel_handle internal_push_container_lookup_fileside(damsel_container file_container_id);

damsel_container container_of_tags_which_are_set_on(damsel_model m, damsel_container c);

damsel_err_t model_map_handles_internal(damsel_container app_container, damsel_container file_container, int late_additions_allowed);

/*! when a container is set to be internal, then it will be released
    automatically when no more references are held to it. */
void set_internal_container(damsel_container c);

/*! when a container has been read from the file, then it should be
    labelled as written so that it does not get written out again if
    we write to this model.
*/
void set_written_container(damsel_container c);


/*! when a container is set the be partially internal, then it will be
    released autoamtically when no more references are to it, but
    if it is a vector container then its memory won't be freed. This is
    for the case where a vector is made from memory passed into an
    API function, but where the user was using (for example) a _fast
    version which in this implementation implicitly makes a container. */
void set_partial_internal_container(damsel_container c);


void model_take(damsel_store_internal *mi);
void model_release(damsel_store_internal *mi);

void DMSLmap_resolve_read_tags_internal(damsel_model model_id);


damsel_err_t DMSLtag_define_internal(damsel_model model_id, damsel_handle_ptr tag_handle, damsel_data_type type, const char *, int written);

#ifdef __cplusplus
}
#endif



#endif

