#ifdef DAMSEL_SUBFILE
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#include <hdf5.h>
#include <hdf5_hl.h>

#ifdef DAMSEL_HDF5_MULTI
#include "h5multidataset.h"
#endif

#include "damsel.h"
#include "damsel-hdf5.h"
#include "damsel-internal.h"
#include "damsel-ndc.h"
#include "damsel-types.h"
#include "damsel-debug.h"
#include "damsel-connectivity.h"
#include "damsel-subfile.h"

typedef struct dset_info_s {
  unsigned quantity;
  char name[DAMSEL_DSET_NAME_MAX]; // what happens when this is overflowed ?
} dset_info;

static MPI_Datatype make_dset_type(dset_info *target) {
    MPI_Datatype newtype;
    int count=0;
    int lens[2] = {sizeof(target->quantity), sizeof(target->name)};
    MPI_Aint offsets[2];
    MPI_Datatype types[2] = {MPI_UNSIGNED, MPI_CHAR};

    MPI_Address(&(target->quantity), &offsets[count++]);
    MPI_Address(&(target->name), &offsets[count++]);

    MPI_Type_create_struct(count, lens, offsets, types,  &newtype);
    MPI_Type_commit(&newtype);

    return newtype;
}

void store_subfiles(damsel_model_internal *s, damsel_attached_file *fs) {
  push_ndc("storing num_subfiles...");

  hsize_t one = 1;
  hid_t dataspace = -1;
  hid_t dataset = -1;
  int ret;

  int myrank;
  MPI_Comm_rank(fs->communicator, &myrank);
  printf_ndc("%s: myrank=%d\n", __func__, myrank);

  if (H5Lexists(fs->file_id, DAMSEL_DATASET_NUM_SUBFILES, H5P_DEFAULT) <= 0) {
    dataspace = H5Screate_simple(1, &one, &one); // single element
    assert(dataspace > 0);

    dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_NUM_SUBFILES,
                        H5T_NATIVE_INT, dataspace, H5P_DEFAULT, 
                        H5P_DEFAULT, H5P_DEFAULT);
    assert(dataset > 0);

    ret = H5Dwrite(dataset, H5T_NATIVE_INT, dataspace, dataspace, H5P_DEFAULT, &s->num_subfiles);
    assert(ret >= 0);

    ret = H5Dclose(dataset);
    assert(ret >= 0);
  } else {
    // need to replace existing value?
    ;
  }

  pop_ndc();
}

/* TODO: most are from store_data. there might be a better to way to get stored chunk name */
void store_data_range(damsel_model_internal *src_s, damsel_attached_file *fs) {
  void *ctx = push_ndc("store_data_range: Writing data range");
  hsize_t one = 1;
  hid_t dataspace = -1;
  hid_t dataset = -1;

  assert(src_s->direction == DAMSEL_TRANSFER_TYPE_WRITE);

  int myrank, processingrank, maxrank;
  myrank = fs->myrank;
  maxrank = fs->maxrank;

  for(processingrank=0; processingrank<maxrank; processingrank++) {
    if(processingrank == myrank) {
      linkedlist *llc;
      for(llc = src_s->first_chunk; llc != NULL; llc=llc->next) {
        dset_info info;
        MPI_Datatype dset_info_type;
        damsel_internal_chunk *chunk = llc->head;
        damsel_container file_container = (damsel_container)(chunk->container);
        damsel_handle desthandle = internal_push_container_lookup_fileside(file_container);
        snprintf(info.name, DAMSEL_DSET_NAME_MAX-1, "%s-%llx-len", chunk->tag->name, desthandle);
        info.quantity = DMSLcontainer_count((damsel_container) chunk->container);
        int mode = 1;
        dset_info_type = make_dset_type(&info);
        MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);
        MPI_Bcast(MPI_BOTTOM, 1, dset_info_type, myrank, fs->communicator);
        hid_t dataspace = H5Screate_simple(1, &one, &one); // single element
        assert(dataspace > 0);
        printf_ndc("master: datasetname=%s\n", info.name);
        dataset = H5Dcreate(fs->file_id, info.name, H5T_NATIVE_INT, 
                            dataspace, H5P_DEFAULT, 
                            H5P_DEFAULT, H5P_DEFAULT);
        assert(dataset > 0);

        int ret;
        ret = H5Dwrite(dataset, H5T_NATIVE_INT, dataspace, dataspace, H5P_DEFAULT, &info.quantity);
        assert(ret >= 0);

        printf("rank(%d): datasetname(%s) is written\n", myrank, info.name);
      
        ret = H5Sclose(dataspace);
        assert(ret >= 0);
        ret = H5Dclose(dataset);
        assert(ret >= 0);
      }
      int mode = 0;
      MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);
    }
    else {
      int mode;
      dset_info info;
      MPI_Datatype dset_info_type = make_dset_type(&info);
      while(1==1) {
        MPI_Bcast(&mode, 1, MPI_INT, processingrank, fs->communicator);
        if(mode==1) {
          MPI_Bcast(MPI_BOTTOM, 1, dset_info_type, processingrank, fs->communicator);
          printf("slave: datasetname=%s\n", info.name);
          hid_t dataspace = H5Screate_simple(1, &one, &one); // single element
          assert(dataspace > 0);
          dataset = H5Dcreate(fs->file_id, info.name, H5T_NATIVE_INT, 
                              dataspace, H5P_DEFAULT, 
                              H5P_DEFAULT, H5P_DEFAULT);
          assert(dataset > 0);

          int ret;
          ret = H5Dwrite(dataset, H5T_NATIVE_INT, dataspace, dataspace, H5P_DEFAULT, &info.quantity);
          assert(ret >= 0);

          printf("rank(%d): datasetname(%s) is written\n", myrank, info.name);
      
          ret = H5Sclose(dataspace);
          assert(ret >= 0);
          ret = H5Dclose(dataset);
          assert(ret >= 0);
        }
        else if (mode == 0) {
          break;
        }
      }
    }
  }
  pop_ndc_context(ctx);
}

#if 0
damsel_err_t damsel_internal_subfile_partition(damsel_model_internal *mi, const char *filename, MPI_Comm comm, damsel_trait_bundle tb) {
  void *ctx = push_ndc("damsel_internal_subfile_partition");

  damsel_err_t err;
  char *mode;
  int i;

  err = DMSLtrait_get(tb, "mode", &mode);
  printf_ndc("mode = %s", mode);

  /* iterate each tag */

  pop_ndc_context(ctx);
  return DMSL_OK;
}
#endif
#endif /* DAMSEL_SUBFILE */
