/* traits code

main user api:
create trait-bundle
add key/value pair
release trait-bundle. causes free if 0 count.

inside damsel:
take reference count (opposite of release)
get access to key/value
  main interface: get a string, given a key.


storage semantics:
trait object has:
  count
  linked list of trait members
    trait members have two string pointers that are malloced specially for this object so can be released when the trait is released. but as long as you have a reference on the trait, you're guaranteed to have it not released.

 */

#include <assert.h>
#include <string.h>

#include "damsel.h"
#include "damsel-internal.h"

#define DAMSEL_TRAIT_MAGIC 58609

typedef struct {
  int magic;
  int count;
  linkedlist *data;
} damsel_trait_bundle_internal;

typedef struct {
  char *key;
  char *value;
} damsel_trait_entry_internal;

damsel_err_t DMSLtrait_create(damsel_trait_bundle *tbp) {
  damsel_trait_bundle_internal *tb = malloc(sizeof(damsel_trait_bundle_internal));
  tb->magic = DAMSEL_TRAIT_MAGIC;
  tb->count = 1;
  tb->data = NULL;

  *tbp = (damsel_trait_bundle) tb;
  return DMSL_OK;
}

void free_trait_entry(void *m) {
  damsel_trait_entry_internal *te = (damsel_trait_entry_internal *)m;
  free(te->key);
  free(te->value);
  free(m);
}

damsel_err_t DMSLtrait_release(damsel_trait_bundle tb) {
  damsel_trait_bundle_internal *t = (damsel_trait_bundle_internal *)tb;
  assert(t->magic == DAMSEL_TRAIT_MAGIC);

  t->count--;

  if(t->count <= 0) {
    assert(t->count == 0);
    linkedlist_free(t->data, free_trait_entry);
    free(tb);
  }
  return DMSL_OK;
}

damsel_err_t DMSLtrait_take(damsel_trait_bundle tb) {
  damsel_trait_bundle_internal *t = (damsel_trait_bundle_internal *)tb;
  assert(t->magic == DAMSEL_TRAIT_MAGIC);
  t->count++;
  return DMSL_OK;
}


/* key and value will be copied, so they can be released by the caller */
damsel_err_t DMSLtrait_put(damsel_trait_bundle tb, char *key, char *value) {
  damsel_trait_bundle_internal *t = (damsel_trait_bundle_internal *)tb;
  assert(t->magic == DAMSEL_TRAIT_MAGIC);

  damsel_trait_entry_internal *te = malloc(sizeof(damsel_trait_entry_internal));

  te->key = malloc(strlen(key) + 1);
  assert(te->key != NULL);
  strcpy(te->key, key);

  te->value = malloc(strlen(value) + 1);
  assert(te->value != NULL);
  strcpy(te->value, value);

  linkedlist_prepend(&(t->data), te);
  return DMSL_OK;
}

/* the string returned here will be deallocated on release of the trait
   bundle. NULL can be passed as parameter, in which case NULL will be
   returned: NULL will be treated as if its an empty trait bundle. */
damsel_err_t
DMSLtrait_get(damsel_trait_bundle tb, char *key, char **value) {
  *value = NULL;
  if(tb == NULL) return DMSL_OK;
  damsel_trait_bundle_internal *t = (damsel_trait_bundle_internal *)tb;
  assert(t->magic == DAMSEL_TRAIT_MAGIC);
  linkedlist *ll;
  for(ll = t->data; ll!=NULL; ll=ll->next) {
    damsel_trait_entry_internal *te = (damsel_trait_entry_internal *) ll->head;
    if (!strcmp(key, te->key)) {
        *value = te->value;
        return DMSL_OK;
    }
  }
  return DMSL_OK;
}

