/* routines for versioning damsel files */

#include <assert.h>
#include <hdf5.h>
#include <sys/time.h>
#include <string.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-hdf5.h"
#include "damsel-ndc.h"
#include "damsel-versioning.h"

/* this is deliberately global; entries are never freed. if this was
   attached to a library instance, then freeing could happen at
   library free time. */
linkedlist *provenance_dependencies = NULL;

hid_t get_versioning_type(damsel_attached_file *file) {
  return file->hdf5_versioning_type;
}

void initialize_versioning_type(damsel_attached_file *file) {
  file->hdf5_versioning_type = create_versioning_type();
}

hid_t create_versioning_type() {
  int ret;
  hid_t versioning_type = H5Tcreate(H5T_COMPOUND,
    sizeof(damsel_hdf5_infile_versioning_watermark));
  assert(versioning_type >= 0);

  ret = H5Tinsert(versioning_type, "collections",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, collections),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "entity_connectivity",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, entity_connectivity),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "seq_handle_containers",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, seq_handle_containers),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "tag_containers",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, tag_containers),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "vector_handle_containers",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, vector_handle_containers),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "timestamp",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, timestamp),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  ret = H5Tinsert(versioning_type, "comm_world_size",
    HOFFSET(damsel_hdf5_infile_versioning_watermark, comm_world_size),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  return versioning_type;
}


hid_t create_versioning_dependency_type() {
  int ret;
  hid_t t = H5Tcreate(H5T_COMPOUND,
    sizeof(damsel_hdf5_infile_versioning_dependency));
  assert(t >= 0);

  ret= H5Tinsert(t, "version",
    HOFFSET(damsel_hdf5_infile_versioning_dependency, version),
    H5T_NATIVE_LONG);
  assert(ret >= 0);

  hid_t string_type = H5Tcopy(H5T_C_S1);
  assert(string_type >= 0);

  ret = H5Tset_size(string_type, MAX_URI_SIZE);
  assert(ret >= 0);

  ret = H5Tinsert(t, "uri",
    HOFFSET(damsel_hdf5_infile_versioning_dependency, uri),
    string_type);

  return t;
}

hsize_t get_extent(damsel_attached_file *fs, char *ds) {

  hsize_t v;
  hid_t d = H5Dopen(fs->file_id, ds, H5P_DEFAULT);
  assert(d >= 0);
  hid_t s = H5Dget_space(d);
  assert(s >= 0);
  v = H5Sget_simple_extent_npoints(s);
  int ret;
  ret = H5Sclose(s);
  assert(ret >= 0);
  ret = H5Dclose(d);
  assert(ret >= 0);

  return v;
}


/* stores all dependencies recorded so far in the runtime
   dependencies list. does not remove entries from the list,
   so they will also be recorded as dependencies of further
   writes in this execution, as well as being transitive
   dependencies. */

void store_dependencies(damsel_model_internal *src_s, damsel_attached_file *fs) {
  int ret;
  const hsize_t hsize_one = 1;
  hsize_t hc_unlimited = H5S_UNLIMITED;
  hid_t dataspace;
  hid_t mem_dataspace;
  hid_t dataset;

  void *ctx = push_ndc("storing provenance dependencies");

  hsize_t nrows = linkedlist_length(provenance_dependencies);

  hid_t dependency_type = create_versioning_dependency_type();
  assert(dependency_type >= 0);

  hsize_t nextrow;

  if(H5Lexists(fs->file_id, DAMSEL_DATASET_VERSIONING_DEPENDENCY, H5P_DEFAULT) <= 0) {
    printf_ndc("creating new provenance dependency dataset");

    printf_ndc("dataset will be created with space for %d entries", nrows);

    dataspace = H5Screate_simple(1, &nrows, &hc_unlimited);
    assert(dataspace >= 0);

    hid_t dependency_properties = H5Pcreate(H5P_DATASET_CREATE);
    assert(dependency_properties >= 0);
    ret = H5Pset_chunk(dependency_properties, 1, &hsize_one);
    assert(ret != 1);

    dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_VERSIONING_DEPENDENCY,
      dependency_type, dataspace, H5P_DEFAULT,
      dependency_properties, H5P_DEFAULT);
    assert(dataset >= 0);
    ret = H5Pclose(dependency_properties);
    assert(ret != 1);
    nextrow = 0;
  } else {
    printf_ndc("extending dataset");

    // otherwise extend it by nrows
    int ret;

    dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_VERSIONING_DEPENDENCY, H5P_DEFAULT);
    assert(dataset >= 0);
    dataspace = H5Dget_space(dataset);
    assert(dataspace >= 0);
    nextrow = H5Sget_simple_extent_npoints(dataspace);
    hsize_t newextent = nextrow+nrows;
    H5Dset_extent(dataset, &newextent);
    printf_ndc("Changing extent to %d", (int) newextent);
    ret = H5Sclose(dataspace);
    assert(ret >= 0);
    dataspace = H5Dget_space(dataset);
    assert(dataspace >= 0);

    ret = H5Sselect_none(dataspace);
    assert(ret >= 0);

    if(nrows > 0) {
      hsize_t *nd = malloc(sizeof(hsize_t) * nrows);
      int i;
      for(i=0; i<nrows; i++) {
        nd[i] = i+nextrow;
        printf_ndc("nd[%d] = %d", i, nd[i]);
      }
      ret = H5Sselect_elements(dataspace, H5S_SELECT_SET, nrows, nd);
      // there may be a way to do this with a range rather than a set but 
      // I don't have hdf5 manual handy
      assert(ret >= 0);
    }
  }

  if(fs->myrank == 0) {
    printf_ndc("rank 0 should store dependency data now");

    damsel_hdf5_infile_versioning_dependency *data = malloc(sizeof(damsel_hdf5_infile_versioning_dependency)* nrows);
    assert(data != NULL);

    printf_ndc("creating mem_dataspace with %d rows", nrows);
    mem_dataspace = H5Screate_simple(1, &nrows, &hc_unlimited);
    assert(mem_dataspace >= 0);

    int pos=0;
    linkedlist *lpos;
    for(lpos = provenance_dependencies; lpos != NULL; lpos = lpos->next) {
      assert(pos < nrows);
      data[pos].version=-1;
      strncpy(data[pos].uri, lpos->head, MAX_URI_SIZE); // BUG this will screw up on URIs longer than MAX_URI_SIZE
      printf_ndc("transcribed %s to %s at %d", lpos->head, data[pos].uri, pos);
      pos++;
    }

    ret = H5Dwrite(dataset, dependency_type, mem_dataspace,
      dataspace, H5P_DEFAULT, data) ;
    assert(ret >= 0);

  } else { // skip on everything other than rank 0
  }

  ret = H5Sclose(dataspace);
  assert(ret >= 0);

  ret = H5Dclose(dataset);
  assert(ret >= 0);

  pop_ndc_context(ctx);

  MPI_Barrier(fs->communicator);
}


/* makes a watermark and stores it in the version table */
void store_version(damsel_model_internal *src_s, damsel_attached_file *fs) {

  const hsize_t hsize_one = 1;
  int res;

  push_ndc("storing version");

  hid_t versioning_type = get_versioning_type(fs);

  hsize_t nextrow;
  hid_t dataspace;
  hid_t dataset;

  // make dataset if it does not exist
  if(H5Lexists(fs->file_id, DAMSEL_DATASET_VERSIONING, H5P_DEFAULT) <= 0 ) {
    hsize_t hc_unlimited = H5S_UNLIMITED;
    dataspace = H5Screate_simple(1, &hsize_one, &hc_unlimited);
    assert(dataspace >= 0);

    hid_t versioning_properties = H5Pcreate(H5P_DATASET_CREATE);
    assert(versioning_properties != -1);
    int ret = H5Pset_chunk(versioning_properties, 1, &hsize_one);
    assert(ret != 1);

    hid_t versioning_type = get_versioning_type(fs);

    dataset = H5Dcreate(fs->file_id, DAMSEL_DATASET_VERSIONING,
      versioning_type, dataspace, H5P_DEFAULT,
      versioning_properties, H5P_DEFAULT);
    assert(dataset >= 0);
    ret = H5Pclose(versioning_properties);
    assert(ret >= 0);
    nextrow=0;
  } else {
    // otherwise extend it by one row
    int ret;

    dataset = H5Dopen(fs->file_id, DAMSEL_DATASET_VERSIONING, H5P_DEFAULT);
    assert(dataset >= 0);
    dataspace = H5Dget_space(dataset);
    assert(dataspace >= 0);
    nextrow = H5Sget_simple_extent_npoints(dataspace);
    hsize_t newextent = nextrow+1;
    H5Dset_extent(dataset, &newextent);
    printf_ndc("Changing extent to %d", (int) newextent);
    ret = H5Sclose(dataspace);
    assert(ret >= 0);
    dataspace = H5Dget_space(dataset);
    assert(dataspace >= 0);
  }

  if(fs->myrank == 0) { // all the versioning stuff happens on rank 0.
    damsel_hdf5_infile_versioning_watermark wm;

    // populate all the fields of wm:

    push_ndc("populating watermark structure");
    wm.collections = get_extent(fs, DAMSEL_DATASET_COLLECTION);
    printf_ndc("collections = %d", wm.collections);
    wm.entity_connectivity = get_extent(fs, DAMSEL_DATASET_CONNECTIVITY);
    printf_ndc("entity_connectivity = %d", wm.entity_connectivity);
    wm.seq_handle_containers = get_extent(fs, DAMSEL_DATASET_CONTAINER_SEQ);
    printf_ndc("seq_handle_containers = %d", wm.seq_handle_containers);
    wm.vector_handle_containers = get_extent(fs, DAMSEL_DATASET_CONTAINER_VECTOR);
    printf_ndc("vector_handle_containers = %d", wm.vector_handle_containers);
    wm.tag_containers = get_extent(fs, DAMSEL_DATASET_TAG);
    printf_ndc("tag_containers = %d", wm.tag_containers);

    struct timeval tv;
    int ret = gettimeofday(&tv, NULL);
    assert(ret == 0);

    wm.timestamp = (long) tv.tv_sec;
    printf_ndc("wm.timestamp = %ld", wm.timestamp);

    int cw_size;
    MPI_Comm_size(MPI_COMM_WORLD, &cw_size); 
    wm.comm_world_size = cw_size;

    pop_ndc();

    // write data
    printf_ndc("writing watermark at row %d", (int) nextrow);

    hid_t hc_plist_id = H5Pcreate(H5P_DATASET_XFER); 
    assert(hc_plist_id >= 0);
    H5Pset_dxpl_mpio(hc_plist_id, H5FD_MPIO_INDEPENDENT);

    hid_t dataspace_one = H5Screate_simple(1, &hsize_one, &hsize_one);
    assert(dataspace_one >= 0);

    res = H5Sselect_none(dataspace);
    assert(res >= 0);

    res = H5Sselect_elements(dataspace, H5S_SELECT_SET, 1, &nextrow);
    assert(res >= 0);
    res = H5Dwrite(dataset, versioning_type, dataspace_one, dataspace,
      hc_plist_id, &wm);
    assert(res >= 0);

    res = H5Sclose(dataspace_one);
    assert(res >= 0);

    res = H5Pclose(hc_plist_id);
    assert(res >= 0);

    // now store dependencies
    // this consists of taking the list of recorded dependencies
    // of which there may be many, and converting them into
    // rows in the dependencies table.
    // there needs to be a structure, either in the library or in
    // the model (I think probably the library)
    // to record these dependencies.

// TODO


  } else { // other rank
  }
  res = H5Sclose(dataspace);
  assert(res >= 0);

  res = H5Dclose(dataset);
  assert(res >= 0);

  pop_ndc();
  MPI_Barrier(fs->communicator);
}

void damsel_provenance_declare_input(char *uri) {
  char *s = malloc(strlen(uri)+1);
  assert(s!=NULL);
  strcpy(s, uri);
  linkedlist_prepend(&provenance_dependencies, uri);
  printf_ndc("declared provenance dependency %s", s);
}

