
/* proto1 - playing around with the API as discussed with quincey so far */

/* you can choose which kind of file store this application will use, by
   either defining or undefining PROTO1_HDF5. If it is undefined, the
   lame text format file store will be used.
*/

#ifdef DAMSEL_HDF5
#define PROTO1_HDF5
#endif


#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include "damsel.h"

#define FILENAME "output/test-proto1.h5"

/* we'll have 7 elements in our rod */
#define LEN 7

double rod[LEN];


void init_sim() {
  int i;
  rod[0] = 99;
  for(i=1; i<LEN; i++) {
    rod[i] = 50;
  }
}

void print_sim() {
  int i;
  printf("* ");
  for(i=0; i<LEN; i++) {
    printf("%.2f ", rod[i]);
  }
  printf("\n");
}

void step_sim() {
  double newstate[LEN];
  int i;
  newstate[0] = rod[0];
  for(i=1; i<LEN-1; i++) {
    newstate[i] = (rod[i-1] + rod[i] + rod[i+1] ) / 3;
  }
  newstate[LEN-1] = rod[LEN-1];
  for(i=0; i<LEN; i++) {
    rod[i] = newstate[i];
  }
}

void dump_sim() {
  printf("dumping\n");

  // initialise in-memory damsel store that maps from other memory
  // but does not contain a copy of that memory (so no mutating of
  // input memory 'till flush)
  damsel_model model;
  damsel_err_t err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);

  // create 'LEN' number of handles, giving a handle container

  // this is DMSLcontainer_create_sequence in the API doc - but its not
  // creating a sequence as much as registering a sequence that was
  // created by the simulation, and making that sequence form a handle
  // container.
  // Also, I changed the parameters a bit.
  damsel_container vertex_container_id;
  err = DMSLcontainer_create_sequence(model,
        (damsel_handle)rod, LEN, sizeof(rod[0]), &vertex_container_id);

  // make them into vertices. whatever that means?
  // what is this call actually supposed to do?
//  errorcode = damsel_define_simple(model, vertex_container, ...);
// skip implementing this for now because i don't have my head around it
// but hopefully i'll understand it later...

  // create a tag for co-ordinates
  // the API has a handle for this tag - so do we use that or the handle?
  // I'm using pointers for my handles. what handle can I use in this case,
  // where the tag doesn't have an in-memory representation in my simulation
  // for me to key off?
  // for now, I can get one easily like this: (ICK!)
  damsel_handle coord_tag_handle = (damsel_handle) malloc(1);

  DMSLtag_define(model, &coord_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "coords");

  short int posns[LEN];
  int i;
  for(i=0;i<LEN; i++) { posns[i] = (short int) (100 * ((float)i)/((float)LEN)); }

  // set the handle values for coord tag to position along 1m rod.
// TODO: I'm unsure if this is the right value for coord_tag_handle...
// I haven't figured out what its really meant to be yet
  DMSLmodel_map_tag(posns, vertex_container_id, &coord_tag_handle);

  // create a tag for temperatures
  // first the same icky mechanism for generating a handle
  damsel_handle temperature_tag_handle = (damsel_handle) malloc(1);

  DMSLtag_define(model, &temperature_tag_handle, DAMSEL_DATA_TYPE_DOUBLE, "temperature");

  // set the handle values for temperature tag to temperatures.
  DMSLmodel_map_tag(rod, vertex_container_id, &temperature_tag_handle);

/*
ok this is the more vague bit...

  // map memory store to file store -- some kind of mapping of handles?
  // sync, flush, close etc.
*/

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  /* this will map the tags to the destination file, using the same
     handles on both sides.
   */
  damsel_handle tags_in_mem[] = {coord_tag_handle, temperature_tag_handle};
  damsel_container tags_container_id;
  err = DMSLcontainer_create_vector(model, tags_in_mem, 2, &tags_container_id);
  DMSLmodel_map_handles_inventing_file_handles(tags_container_id);

  DMSLmodel_map_handles_inventing_file_handles(vertex_container_id);

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);

  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  DMSLmodel_close(model);
  DMSLcontainer_close(tags_container_id);
  DMSLcontainer_close(vertex_container_id);

  free((void *)coord_tag_handle);
  free((void *)temperature_tag_handle);

  printf("done dumping\n");
}

void reload_sim() {
  // damsel_store *filestore = DMSLstore_open_hdf5(FILENAME);
  /* my assumption is that tags are already defined in the filestore, as
     will be handle containers later; but that I need to create them in
     the memory store. */

  damsel_model model;
  damsel_err_t err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_populate(model);

  /* the following two define a tag and a vertex container
     as in the dump routine. the handles don't need to match up. */
  damsel_handle temperature_tag_handle = (damsel_handle) malloc(1);
  DMSLtag_define(model, &temperature_tag_handle, DAMSEL_DATA_TYPE_DOUBLE, "temperature");

  damsel_container vertex_container_id;
  err = DMSLcontainer_create_sequence(model,
        (damsel_handle)rod, LEN, sizeof(rod[0]), &vertex_container_id);

  /* now I have a get call, which mirrors set.
     at the time of making this call, I don't have anything that specifies
     *where* to get the data from - just "the memory store" which never
     actually holds anything. */
  DMSLmodel_map_tag(rod, vertex_container_id, &temperature_tag_handle);

  damsel_handle fileside_temperature_tag_handle;
  err = DMSLselect_tag_by_name(model, "temperature", &fileside_temperature_tag_handle);

  damsel_container filestore_vertex_container_id = DMSLselect_handles_with_values(model, fileside_temperature_tag_handle);

  DMSLmodel_map_handles(vertex_container_id, filestore_vertex_container_id);
  DMSLcontainer_release(filestore_vertex_container_id);


  /* note that this is an execute on model not on filestore, because of
     the way that I've implemented store_map: you need to execute on the
     place data is going, not coming from. that doesn't have to be the
     case - I was just too lazy to implement that... */
  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);

  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  DMSLmodel_close(model);
  DMSLcontainer_close(vertex_container_id);

  free((void *)temperature_tag_handle);

}

int main(int argc, char **argv) {
  printf("start proto1\n");

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  printf("MPI initialized\n");

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  if(argc == 2 && !strcmp("continue",argv[1])) {
    reload_sim();
  } else {
    init_sim();
  }


  print_sim();

  int step = 1;
  while(step%10 != 0) {
    printf("simulation step %d\n", step);
    step_sim();
    print_sim();
    step++;
  }

  printf("done simulating. now dumping.\n");

  dump_sim();

  DMSLlib_finalize(lib);
  MPI_Finalize();

  printf("end proto1\n");
  return 0;
}

