/* 

A scaling test based on test-mpi.c

I would like to make it like this:

One parameter is the number of ranks.
Another parameter is the data set siz.
Another parameter is the write pattern:
   test-mpi.c does round-robin. (A)
   Also would like to put each rank's
     data contiguously (B).
   The contiguous  form will give a
   different, simpler mapping pattern,
   I think.
*/

#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>

#include "mpi.h"
#include "damsel.h"
#include "damsel-debug.h"

#ifdef HAVE_GPERFTOOLS_PROFILER_H
#include <gperftools/profiler.h>
#endif
#ifdef HAVE_GOOGLE_PROFILER_H
#include <google/profiler.h>
#endif

#define FILENAME "output/scale-mpi.h5"

int *proc_ranks; // = {0,0,0,0,0};
int tagrank = 0;

int maxrank, myrank;

struct timeval starttv;

void log(char *s) {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  printf("%d.%6.6d %d %s\n", tv.tv_sec, tv.tv_usec, myrank, s);
}

void start() {
  printf("Starting\n");
  gettimeofday(&starttv, NULL);
}

void end() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  long startus = starttv.tv_sec * 1000000 + starttv.tv_usec;
  long endus = tv.tv_sec * 1000000 + tv.tv_usec;
  long td = endus - startus;
  printf("Ended rank %d. time difference = %ld us\n", myrank, td);
}

int main(int argc, char **argv) {
#ifndef DAMSEL_SINGLETHREAD
  int res;
#endif
  short int *data; // = {3, 1, 4, 1, 5}; // some arbitrary data to write
  int i;

  int maxd;
  maxd=atoi(argv[1]);


  printf("data size = %d\n", maxd);

  proc_ranks = malloc(sizeof(int) * maxd);
  data = malloc(sizeof(short int) * maxd);

  printf("%s\n", argv[0]);

#ifndef DAMSEL_SINGLETHREAD
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif
  log("non-damsel prep");
  MPI_Comm_size(MPI_COMM_WORLD, &maxrank);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

#ifdef HAVE_LIBPROFILER
  char profile_output[128];
  sprintf(profile_output,"./scale-mpi-%i.prof",myrank);
  ProfilerStart(profile_output);
#endif

  /* we'll create a vertex handle container with my ranks handles in it. */
  int handles_on_this_rank = 0;
  damsel_handle *h; //[5]
  h = malloc(sizeof(damsel_handle) * maxd);
  short int *mydata; // [5]
  mydata = malloc(sizeof(damsel_handle) * maxd);
  
  for(i=0;i<maxd;i++) {
    if(i % maxrank == myrank) {
      h[handles_on_this_rank] = 0x9990 + i;
      mydata[handles_on_this_rank] = data[i];
      handles_on_this_rank++;
    }
  }

  printf("rank %d will output %d handles.\n", myrank, handles_on_this_rank);

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently ignore any error

  log("Initializing library");
  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  log("Creating model");
  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  /* define a tag which will store our data. the tag will be
     named 'somedata' */

  log("Defining tag");
  damsel_handle data_tag_handle = 0x10203;
  DMSLtag_define(model, &data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "somedata");


  log("Creating vector");
  damsel_container vertex_container_id;
  err = DMSLcontainer_create_vector(model, h, handles_on_this_rank, &vertex_container_id);
  //DMSLmodel_map_handles_inventing_file_handles(vertex_container_id);
    // another test would be to map out the handles for the whole container

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(mydata, vertex_container_id, &data_tag_handle);

  /* at this point, the in-memory store knows about all the data that
     we have. next we will write the data out to disk.
     we need to invent some handles for the file store too. They can overlap
     with memstore handles, though here I make sure that they do not.
  */

  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  log("Attaching");
  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  // now I'm using the same handle values on all ranks.
  // so map everything to the handle space on the first rank.

  damsel_err_t ret;
  if(myrank != 0) {
    // this should map everything on a rank into the same handles on the
    //  0 rank, which should then do the writing.
    log("Global handles map");
    ret =  DMSLmodel_map_global_handles(vertex_container_id, proc_ranks, h);
    assert_dmsl_ok(ret);

    damsel_container tag_container;
    err = DMSLcontainer_create_sequence(model, data_tag_handle, 1, 1, &tag_container);

    ret = DMSLmodel_map_global_handles(tag_container, &tagrank, &data_tag_handle);
    assert_dmsl_ok(ret);
  }


  /* we also need to map the handles for the tags which contain values
     that we want to store.
     This happens by creating a collection containing the tag to map, and
     mapping that into the file store. */

  if(myrank == 0) DMSLmodel_map_single_handle_inventing_file_handle(model, data_tag_handle);

  damsel_container full_vertex_container_id;
  err = DMSLcontainer_create_sequence(model, 0x9990, maxd, 1, &full_vertex_container_id);

  if(myrank == 0) {
    damsel_handle collection_handle = 0x333;
    DMSLcoll_create(model, &collection_handle, full_vertex_container_id, DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR);
    DMSLmodel_map_single_handle_inventing_file_handle(model, collection_handle);
    DMSLmodel_map_handles_inventing_file_handles(full_vertex_container_id);
  }



  /* Now we can tell damsel to execute everything. */

  log("Transfer");
  start();
  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);
  end();
  log("Close file");
  DMSLmodel_close(model);
  log("Non-time related tidyup");
  DMSLcontainer_close(full_vertex_container_id);
  DMSLcontainer_close(vertex_container_id);

  DMSLlib_finalize(lib);

  MPI_Finalize();
#ifdef HAVE_LIBPROFILER
  ProfilerStop();
#endif
  return 0;
}


