/*

  test 3d entity writing

  there will be 6 cubes, in a 2x3x1 layout.

      A - B - C - D
      |   |   |   |
      E - F - G - H
      |   |   |   |
      I - J - K - L

(and in the other, depth, dimension A - A' etc)

The six cubes are labelled u v w
                           x y z

I want to give these handles in a permuted fashion so that any
rearrangements made by the connectivity code can be seen to be
happening. The app side handles will be in the range 4000 - 4999

I will store 24 numbers from 2000 to 2023 on a tag on the vertices,
in left->right, top->bottom, front->back:  1  2  3  4     \   13 14 15 16
                                           5  6  7  8     \   17 18 19 20
                                           9 10 11 12     \   21 22 23 24

I will store 6 numbers from 50-55 on a tag on the cubes:

numbers:
   50 51 52
   53 54 55


Handles for the vertices:

   4010  4120  4040  4230
   4190  4100  4220  4070
   4210  4160  4050  4140

    \       \      \  \    \

       4060  4170  4150  4090
       4080  4240  4200  4180
       4030  4130  4110  4020

Handles for the cubes:
       4501  4504  4502
       4503  4505  4500

Connectivity: For each cube, I need to list out the handles of its vertices,
using the above tables. This can then be described to damsel in the
entity definition call.
*/

#include <assert.h>
#include <stdio.h>
#include <unistd.h>

#include "mpi.h"
#include "damsel.h"

#define FILENAME "output/test-connectivity-write-cubes-ord.h5"

int main(int argc, char **argv) {
#ifndef DAMSEL_SINGLETHREAD
  int res;
#endif

// handles for the vertices:
//   4010  4020  4030  4040  4050  4060  4070  4080  4090  4100  4110  4120

  short int cube_data[6] = {55, 50, 52, 53, 51, 54};

  short int vertex_data[24] = {
      1,    24,  21,    3,   11,    13,  8,    17,    16,   6,   23,   2,

//   4130  4140  4150  4160  4170  4180  4190  4200  4210  4220  4230  4240
      22,   12,   15,   10,   14,   20,   5,    19,   9,    7,    4,    18
  };

  // as there is no fixed ordering for vertices (or even for the number of
  // vertices for each type of entity) in the damsel API, I'll put this in
  // in rather arbitrary order.
  damsel_handle connectivity[] = {
    4220, 4070, 4050, 4140, 4200, 4180, 4110, 4020, // cube 4500
    4010, 4120, 4190, 4100, 4060, 4170, 4080, 4240, // cube 4501
    4040, 4230, 4220, 4070, 4150, 4090, 4200, 4180, // cube 4502
    4190, 4100, 4210, 4160, 4080, 4240, 4030, 4130, // cube 4503
    4120, 4040, 4100, 4220, 4170, 4150, 4240, 4200, // cube 4504
    4100, 4220, 4160, 4050, 4240, 4200, 4130, 4110, // cube 4505
  };

  printf("%s\n", argv[0]);

#ifndef DAMSEL_SINGLETHREAD
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  /* define a tag which will store our data. the tag will be
     named 'somedata' */

  damsel_handle vertex_data_tag_handle = 0x4001;
  DMSLtag_define(model, &vertex_data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "vertexdata");

  damsel_handle cube_data_tag_handle = 0x4002;
  DMSLtag_define(model, &cube_data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "cubedata");

  /* we'll create a sequence handle container with 5 handles in it. */

  damsel_container vertex_handles_container_id;
  err = DMSLcontainer_create_sequence(model, 4010, 24, 10, &vertex_handles_container_id);

  damsel_container cube_handles_container_id;
  err = DMSLcontainer_create_sequence(model, 4500, 6, 1, &cube_handles_container_id);

  DMSLmodel_map_tag(vertex_data, vertex_handles_container_id, &vertex_data_tag_handle);
  DMSLmodel_map_tag(cube_data, cube_handles_container_id, &cube_data_tag_handle);

  DMSLentity_define_fast(cube_handles_container_id, DAMSEL_ENTITY_TYPE_HEX, 8, connectivity);

  /* at this point, the in-memory store knows about all the data that
     we have. next we will write the data out to disk.
     we need to invent some handles for the file store too. They can overlap
     with memstore handles, though here I make sure that they do not.
  */

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently ignore any error
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");
  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_map_handles_inventing_file_handles_with_permute(vertex_handles_container_id, damsel_permute_connectivity);
  DMSLmodel_map_handles_inventing_file_handles_with_permute(cube_handles_container_id, damsel_permute_connectivity);
  // DMSLmodel_map_handles_inventing_file_handles(cube_handles_container_id);

  /* we also need to map the handles for the tags which contain values
     that we want to store.
     This happens by creating a container containing the tag to map, and
     mapping that into the file store. */

  DMSLmodel_map_single_handle_inventing_file_handle(model, vertex_data_tag_handle);
  DMSLmodel_map_single_handle_inventing_file_handle(model, cube_data_tag_handle);

  /* Now we can tell damsel to execute everything. */

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);


  DMSLmodel_close(model);
  model=NULL;

  DMSLcontainer_close(cube_handles_container_id);
  cube_handles_container_id=NULL;

  DMSLcontainer_close(vertex_handles_container_id);
  vertex_handles_container_id=NULL;

  DMSLlib_finalize(lib);
  lib=NULL;

  MPI_Finalize();
  return 0;
}

