
#include <assert.h>
#include <mpi.h>
#include "damsel.h"
#include "damsel-hdf5.h"
#include "damsel-connectivity.h"

int main(int argc, char **argv) {

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res>=MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  assert(argc==2);
  char *filename = argv[1];

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);
  damsel_model_internal *m_model = (damsel_model_internal *) model;

  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");
  DMSLmodel_attach(model, filename, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  DMSLmodel_populate(model);

  int seq_container_count, tag_count, container_count,
    vec_row_count;
  DMSLmodel_get_tuple_count(model, &container_count, &tag_count);
  DMSLhdf5_get_container_counts(m_model->attached_file, &seq_container_count, &vec_row_count);

  printf("File: %s\n", filename);
  printf("Container count: %d\n", container_count);
  printf("Sequence container count: %d\n", seq_container_count);
  printf("Vector container rows: %d\n", vec_row_count);
// TODO: this assert is wrong because it talks about vector rows not containers
  // assert(container_count == (seq_container_count + vec_container_count));

  printf("Tag count: %d\n", tag_count);

  damsel_tag_buf_type *col =
    malloc(tag_count * sizeof(damsel_tag_buf_type));

  DMSLmodel_get_tag_infos(model, col);

  printf("===== Tag definitions =====\n");
  int tag_num;
  for(tag_num=0; tag_num<tag_count; tag_num++) {
    printf("Tag %d:\n", tag_num);
    printf("  name: %s\n", col[tag_num].name);
    printf("  type: %lld\n", col[tag_num].tag_datatype);
    printf("handle: %llx\n", *(col[tag_num].tag_handle));
  }
  printf("===== end of tag definitions =====\n");
  free(col);

  damsel_container_buf_type *container_buf =
    malloc(container_count * sizeof(damsel_container_buf_type));

  DMSLmodel_get_container_infos(model, container_buf);
  int container_num;
  for(container_num=0; container_num < container_count; container_num++) {
    printf("Container %d:\n", container_num);
    printf("  id: %p\n", container_buf[container_num]);
    damsel_container_internal *hc = (damsel_container_internal *)container_buf[container_num];
    switch(hc->type) {
      case DAMSEL_HANDLE_CONTAINER_TYPE_SEQUENCE:
        printf("  type: sequence\n");
        break;
      case DAMSEL_HANDLE_CONTAINER_TYPE_VECTOR:
        printf("  type: vector\n");
        break;
      default:
        printf("  type: unknown\n");
        break;
    }
  }
  free(container_buf);

  int entity_count;
  entity_count = DMSLmodel_get_entity_count(model);


  if(entity_count != 0) {
    damsel_entity_buf_type *entity_buf =
      malloc(entity_count * sizeof(damsel_entity_buf_type));
    DMSLmodel_get_entity_infos(model, entity_buf);

    printf("===== Entity definitions =====\n");
    printf("Entity definition count: %d\n", entity_count);


    int entity_num;
    for(entity_num = 0; entity_num < entity_count; entity_num++) {
      printf("Entity definition %d\n", entity_num);
      printf(" type %d\n", entity_buf[entity_num].entity_type);
      printf(" vertices per entity %zd\n", entity_buf[entity_num].vertices_per_entity);
      printf(" count %zd\n", entity_buf[entity_num].count);
      printf(" entity container id %p\n", entity_buf[entity_num].entity_container);
      printf(" vertex container id %p\n", entity_buf[entity_num].vertex_container);
      printf(" tag count %zd\n", entity_buf[entity_num].tag_count);
      printf(" tag container id %p\n", entity_buf[entity_num].tag_handle_container);
      // TODO: more informationc could be given about the contents of
      // the three above containers.
    }
    printf("===== end of entity definitions =====\n");
    DMSLmodel_close_entity_infos(model, entity_buf);
    free(entity_buf);
  } else {
    printf("This file contains no entity definitions\n");
  }


  int collection_count;

  collection_count = DMSLmodel_get_collection_count(model);
  damsel_collection_buf_type *col_buf = NULL;
  if(collection_count != 0) {
    col_buf = malloc(collection_count * sizeof(damsel_collection_buf_type));
    DMSLmodel_get_collection_infos(model, col_buf);
  }

  printf("===== Collection definitions =====\n");
  printf("Collection count: %d\n", collection_count);
  int collection_num;
  for(collection_num = 0; collection_num < collection_count ; collection_num++) {
    printf("Collection definition %d\n", collection_num);
    printf("type %d\n", col_buf[collection_num].type);
    printf("fileside handle 0x%llx\n", *(col_buf[collection_num].collection_handle));
    printf("tag container %p\n", col_buf[collection_num].tag_handle_container);
    printf("tag count %zd\n", col_buf[collection_num].tag_count);
    printf("contents container %p\n", col_buf[collection_num].contents);
  }
  printf("===== end of collection definitions =====\n");

  DMSLmodel_close_collection_infos(model, col_buf);
  free(col_buf);

  DMSLmodel_close((damsel_model)m_model);
  DMSLlib_finalize(lib);
  MPI_Finalize();

  exit(0);
}
