#include <assert.h>
#include <stdio.h>
#include <time.h>

#include "damsel.h"
#include "damsel-sugar.h"
#include "damsel-internal.h"
#include "hdf5.h"

#define FNAME "output/test-mem-leak_2.h5"
#define NCONTS 2
#define NPOINTS 40000000

#ifndef DBL_EPSILON 
#define DBL_EPSILON 1E-9
#endif

int test_write(const char **names, const float *data, int n, damsel_handle *handles) ;
int test_read(const char *names, const float *targetdata, int n);
int test_write_read(int n, int isVector);
int test_write_loop(int n, int isVector);
int test_mp();
int test_mp_indiv(int ndsets, int npoints);
int test_mp_multi(int ndsets, int npoints);
	
int main(int argc, char **argv) 
{
    int nloops=0, npoints=NPOINTS, i, isVector=0, isMP=0;
	
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &i);

	/**
	    -l N:	number of loops
		-n N:	number of data points
		-v  :	test for vector container
		-m  :   test for MP
	 */
	for (i=0; i<argc; i++) {
	    if (strcmp(argv[i], "-l")==0)
		    nloops = atoi(argv[++i]);
	    else if (strcmp(argv[i], "-n")==0)
		    npoints = atoi(argv[++i]);
	    else if (strcmp(argv[i], "-v")==0)
		    isVector = 1;
	    else if (strcmp(argv[i], "-m")==0)
		    isMP = 1;
	}
		
	if (nloops>0)
	    test_write_loop(nloops, isVector);
	else if (isMP)
	    test_mp();
	else
	    test_write_read(npoints, isVector);
		
    MPI_Finalize();
}

int test_mp()
{
    int ndsets=20, npoints=NPOINTS, myrank;

    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	
    printf("No. of datasets = %d, \t No. of data points = %d\n", ndsets, npoints);
    
    printf("Rank%d: Time with individual writes = %d\n", myrank, test_mp_indiv(ndsets, npoints));
    printf("Rank%d: Time with multidataset writes = %d\n", myrank, test_mp_multi(ndsets, npoints));
	
	return 0;
}	

int test_mp_multi(int ndsets, int npoints)
{
    int i, retval=0, err=0;
	hid_t fid=-1, fsid=-1, dids[80], mpiprop=-1, xprop=H5P_DEFAULT;
	hsize_t dims=npoints;
	int *rdata=NULL, *wdata=NULL, myrank, maxrank, count, nbytes, displacement[80];
	char dset_name[80];
	haddr_t offsets[80];
	MPI_File fh;
    MPI_Datatype type;	
	MPI_Status status;
	
	if (ndsets>80) ndsets=80;

	MPI_Comm_size(MPI_COMM_WORLD, &maxrank);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	mpiprop = H5Pcreate (H5P_FILE_ACCESS);
    H5Pset_fapl_mpio(mpiprop, MPI_COMM_WORLD, MPI_INFO_NULL);
	
    xprop = H5Pcreate (H5P_DATASET_XFER);
    H5Pset_dxpl_mpio (xprop, H5FD_MPIO_COLLECTIVE);
  
	fid = H5Fcreate (FNAME, H5F_ACC_TRUNC, H5P_DEFAULT, mpiprop);
	fsid = H5Screate_simple (1, &dims, NULL);
	
	for (i=0; i<ndsets; i++) {
	    sprintf(dset_name, "dset%d", i);
		dids[i] = H5Dcreate (fid, dset_name, H5T_NATIVE_INT, fsid, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
		offsets[i] = H5Dget_offset(dids[i]); 
	}

    MPI_Bcast(offsets, ndsets*sizeof(haddr_t), MPI_BYTE, myrank, MPI_COMM_WORLD);

	H5Pclose(xprop);	
	H5Pclose(mpiprop);	
	H5Sclose(fsid);
	for (i=0; i<ndsets; i++)
	    H5Dclose(dids[i]);	
	H5Fclose(fid);	
	
	count = npoints;
	nbytes = count*sizeof(int);
	wdata = (int *)malloc(nbytes);
	for (i=0; i<count; i++)
	    wdata[i] = myrank+1;
	
	for (i=0; i<ndsets; i++)
	    displacement[i] = offsets[i]+myrank*count;
		
    MPI_Type_create_indexed_block(ndsets, nbytes, displacement, MPI_BYTE, &type);
    MPI_Type_commit(&type);
	
    err = MPI_File_open(MPI_COMM_WORLD, FNAME, MPI_MODE_RDWR, MPI_INFO_NULL, &fh);
	//printf("error code from MPI_File_open() = %d\n", err);
	
    err = MPI_File_set_view(fh, 0, MPI_BYTE, type, "native", MPI_INFO_NULL);
	//printf("error code from MPI_File_set_view() = %d\n", err);
	
    MPI_Barrier(MPI_COMM_WORLD);

  	time_t t0 = time(NULL);
	err = MPI_File_write_all( fh, wdata, nbytes, MPI_BYTE, &status );
	time_t t1 = time(NULL);
	printf("error code from MPI_File_write_all() = %d\n", err);

    MPI_Barrier(MPI_COMM_WORLD);	
	free(wdata);
    MPI_File_close(&fh);

	return  (int) (t1-t0);
}

int test_mp_indiv(int ndsets, int npoints)
{
    int i, retval=0;
	hid_t fid=-1, fsid=-1, dids[80], mpiprop=-1, xprop=H5P_DEFAULT;
	hsize_t dims=npoints;
	int *rdata=NULL, *wdata=NULL, myrank;
	char dset_name[80];

	if (ndsets>80) ndsets=80;
	
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	mpiprop = H5Pcreate (H5P_FILE_ACCESS);
    H5Pset_fapl_mpio(mpiprop, MPI_COMM_WORLD, MPI_INFO_NULL);
	
    xprop = H5Pcreate (H5P_DATASET_XFER);
    H5Pset_dxpl_mpio (xprop, H5FD_MPIO_COLLECTIVE);
  
	fid = H5Fcreate (FNAME, H5F_ACC_TRUNC, H5P_DEFAULT, mpiprop);
	fsid = H5Screate_simple (1, &dims, NULL);
	
	for (i=0; i<ndsets; i++) {
	    sprintf(dset_name, "dset%d", i);
		dids[i] = H5Dcreate (fid, dset_name, H5T_NATIVE_INT, fsid, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	}

	wdata = (int *)malloc(npoints*sizeof(int));
	for (i=0; i<npoints; i++)
	    wdata[i] = myrank;
	
    MPI_Barrier(MPI_COMM_WORLD);

  	time_t t0 = time(NULL);
	for (i=0; i<ndsets; i++)
        H5Dwrite (dids[i], H5T_NATIVE_INT, H5S_ALL, fsid, xprop, wdata);
	time_t t1 = time(NULL);

	free(wdata);

    MPI_Barrier(MPI_COMM_WORLD);
	H5Pclose(xprop);	
	H5Pclose(mpiprop);	
	H5Sclose(fsid);
	for (i=0; i<ndsets; i++)
	    H5Dclose(dids[i]);	
	H5Fclose(fid);

	return  (int) (t1-t0);
}

/*
int test_mp_indiv(int ndsets, int npoints)
{
    int i, retval=0;
	hid_t fid=-1, msid=-1, fsid=-1, did=-1, mpiprop=-1, xprop=H5P_DEFAULT;
	hsize_t dims=npoints, start, count;
	int *rdata=NULL, *wdata=NULL;
	int maxrank, myrank;

    MPI_Comm_size(MPI_COMM_WORLD, &maxrank);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	mpiprop = H5Pcreate (H5P_FILE_ACCESS);
    H5Pset_fapl_mpio(mpiprop, MPI_COMM_WORLD, MPI_INFO_NULL);
	
    xprop = H5Pcreate (H5P_DATASET_XFER);
    H5Pset_dxpl_mpio (xprop, H5FD_MPIO_COLLECTIVE);
  
	fid = H5Fcreate (FNAME, H5F_ACC_TRUNC, H5P_DEFAULT, mpiprop);
	fsid = H5Screate_simple (1, &dims, NULL);
	did = H5Dcreate (fid, "data", H5T_NATIVE_INT, fsid, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	count = NPOINTS/maxrank;
	start = myrank*count;
	msid = H5Screate_simple (1, &count, NULL);
		
    MPI_Barrier(MPI_COMM_WORLD);

	wdata = (int *)malloc(count*sizeof(int));
	for (i=0; i<count; i++) wdata[i] = myrank+1;
	H5Sselect_hyperslab (fsid, H5S_SELECT_SET, &start, NULL, &count, NULL);

  	time_t t0 = time(NULL);
    H5Dwrite (did, H5T_NATIVE_INT, msid, fsid, xprop, wdata);
	time_t t1 = time(NULL);

	free(wdata);

    MPI_Barrier(MPI_COMM_WORLD);
	H5Pclose(xprop);	
	H5Pclose(mpiprop);	
	H5Sclose(msid);
	H5Sclose(fsid);
	H5Dclose(did);	
	H5Fclose(fid);

	printf("############ time=%d, maxrank=%d, myrank=%d\n", (t1-t0), maxrank, myrank, fid, msid, fsid, did);fflush(stdout);

	return  (int) (t1-t0);
}
*/

int test_write_loop(int nloops, int isVector)
{
    int i,j, n=2;
	float *data=NULL;
	damsel_handle *handles=NULL;
	char *names[NCONTS] = {"data1", "data2"};
	
	for (i=0; i<nloops; i++)
	{
	    n = n<<1;
        data = (float *)malloc(n*sizeof(float));
	    for (j=0; j<n; j++) data[j] = (float)j;
		
		if (isVector) {
          handles = (damsel_handle *)malloc(n*sizeof(damsel_handle));
	      for (j=0; j<n; j++) handles[j] = (damsel_handle)(0x200 + 0x20*j);
		}
		
	    printf("npoints=%d\t time=%d\n", NCONTS*n, test_write(names, data, n, handles));

	    free (data);
		if (handles) {
		  free(handles);
		  handles = NULL;
		}
	}
        	

	return 0;
}

int test_write_read(int n, int isVector)
{
    int i, retval=0;
	float *data=NULL;
	damsel_handle *handles=NULL;
	
	char *names[] = {"data1", "data2"};

    data = (float *)malloc(n*sizeof(float));
	for (i=0; i<n; i++)
	    data[i] = (float)1.0/(float)(i+1.0);

	if (isVector) {
      handles = (damsel_handle *)malloc(n*sizeof(damsel_handle));
	  for (i=0; i<n; i++) handles[i] = (damsel_handle)(0x200 + 0x20*i);
	}
		
	retval = test_write(names, data, n, handles);
	if (retval < 0 || retval>60)
	    printf("The writing test failed, It took %ld seconds.\n", retval);
    else printf("Writing passed. It took %d seconds\n", retval);
	
	// retval = test_read(names[1], data, n);
	// if (retval < 0 || retval>60)
	    // printf("The reading test failed, It took %ld seconds.\n", retval);
    // else printf("Reading passed. It took %d seconds\n", retval);
	
done:	
	free (data);
	if (handles) {
      free(handles);
	  handles = NULL;
	}	
	return retval;
}
	
int test_write(const char *names[], const float *data, int n, damsel_handle *handles) 
{
    int i;
    damsel_library lib;
    damsel_err_t err = DMSLlib_init(&lib);
    damsel_handle data_tag_handles[] = {100, 101};
	damsel_container cont_ids[NCONTS];
	damsel_handle *vector_of_handles;
	
    unlink(FNAME);
    damsel_model model = DMSLmodel_attached_with_trait(MPI_COMM_WORLD, FNAME, "WRITE");
  
    for (i=0; i<NCONTS; i++) {
      DMSLtag_define(model, &data_tag_handles[i], DAMSEL_DATA_TYPE_FLOAT, names[i]);
	
	  if (handles)
	    err = DMSLcontainer_create_vector(model, handles, n, &cont_ids[i]);
	  else
        err = DMSLcontainer_create_sequence(model, 0x200, n, 0x20, &cont_ids[i]);
	  DMSLmodel_map_tag(data, cont_ids[i], &data_tag_handles[i]);
	
      DMSLmodel_map_handles_inventing_file_handles(cont_ids[i]);
      DMSLmodel_map_single_handle_inventing_file_handle(model, data_tag_handles[i]);
	  DMSLcontainer_release(cont_ids[i]);
	}

	time_t t0 = time(NULL);
    DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_WRITE);
	time_t t1 = time(NULL);
	
    DMSL_CHECK(DMSLmodel_close(model));
    DMSLlib_finalize(lib);
	
    return (int) (t1-t0);
}

int test_read(const char *name, const float *targetdata, int n) 
{
    int i, retval=0;
	float                  *data;
    damsel_handle          data_tag_handle = 0x500;
    damsel_library lib;
    damsel_err_t err = DMSLlib_init(&lib);
    damsel_handle          filefside_data_tag_handle;
    damsel_container       filestore_vertex_collection_id;
	damsel_container       file_tag_collection_id;
	damsel_container       tag_collection_id;
	damsel_container       vertex_collection_id;
    damsel_model           model = DMSLmodel_attached_with_trait(MPI_COMM_WORLD, FNAME, "READ");
    damsel_model_internal *ms = (damsel_model_internal *)model;
  
    DMSL_CHECK(DMSLmodel_populate(model));

  /* set up filefside things */
  err = DMSLselect_tag_by_name(model, name, &filefside_data_tag_handle);
  err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file,
        filefside_data_tag_handle, 1, 1, &file_tag_collection_id);
  filestore_vertex_collection_id = DMSLselect_handles_with_values(
	  model, filefside_data_tag_handle);
  
  /* set up memfside things */
  err = DMSLcontainer_create_sequence(model, 0x200, n, 0x20, &vertex_collection_id);
  err = DMSLcontainer_create_sequence(model, data_tag_handle, 1, 1, &tag_collection_id);	  
 
  /* mapping memfside to filefside */
  DMSLmodel_map_handles(tag_collection_id, file_tag_collection_id);
  DMSLmodel_map_handles(vertex_collection_id, filestore_vertex_collection_id);

  data = (float *)malloc(n*sizeof(float));
  DMSLmodel_map_tag(data, vertex_collection_id, &data_tag_handle);

  time_t t0 = time(NULL);
  DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_READ);
  time_t t1 = time(NULL);
	
  DMSLcontainer_release(file_tag_collection_id);
  DMSLcontainer_release(filestore_vertex_collection_id);
  DMSLcontainer_release(tag_collection_id);
  DMSLcontainer_release(vertex_collection_id);

  DMSL_CHECK(DMSLmodel_close(model));
  DMSLlib_finalize(lib);
	
	for (i=0; i<n; i++) {
		if ((data[i]-targetdata[i])>DBL_EPSILON) {
			retval = -1;
			goto done;
		}
	}
	
	retval = (int) (t1-t0);
		
done:		
	free (data);
    return retval;
}
