/* 
 * test-pnetcdf.c:  extract the pnetcdf-on-damsel work into a separate
 * standalone testcase.  Goal is to exercise damsel pieces pnetcdf needs to use
 * while lowering barrier of entry for other eyeballs
 *
 * pnetcdf/netcdf has the following main features
 * - dimensions
 * - variables
 * - attributes
 * - data
 *
 * A simple pnetcdf example would define a dimension, define a few variables
 * using that dimension, then write, in parallel, those variables.
 */

#include "mpi.h"
#include "damsel.h"
#include "damsel-internal.h"

#include "damsel-sugar.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <stdint.h>

#include <stddef.h>

typedef enum {
        NC_NAT =        0,      /* NAT = 'Not A Type' (c.f. NaN) */
        NC_BYTE =       1,      /* signed 1 byte integer */
        NC_CHAR =       2,      /* ISO/ASCII character */
        NC_SHORT =      3,      /* signed 2 byte integer */
        NC_INT =        4,      /* signed 4 byte integer */
        NC_FLOAT =      5,      /* single precision floating point number */
        NC_DOUBLE =     6,      /* double precision floating point number */
        NC_TYPE_ENUM_SIZE
} nc_type;

#define NC_NOERR        0       /* No Error */


static int damsel_type_to_netcdf_type(damsel_data_type type) 
{
    switch (type) {
	case DAMSEL_DATA_TYPE_BYTES:
	    return NC_BYTE;
	case DAMSEL_DATA_TYPE_SHORTINT:
	    return NC_SHORT;
	case DAMSEL_DATA_TYPE_INTEGER:
	    return NC_INT;
	case DAMSEL_DATA_TYPE_FLOAT:
	    return NC_FLOAT;
	case DAMSEL_DATA_TYPE_DOUBLE:
	    return NC_DOUBLE;
	default:
	    return NC_NAT;
    }
}

static int netcdf_type_to_damsel_type(nc_type type) {
    switch (type) {
	case NC_BYTE:
	    return DAMSEL_DATA_TYPE_BYTES;
	case NC_SHORT:
	    return DAMSEL_DATA_TYPE_SHORTINT;
	case NC_INT:
	    return DAMSEL_DATA_TYPE_INTEGER;
	case NC_FLOAT:
	    return DAMSEL_DATA_TYPE_FLOAT;
	case NC_DOUBLE:
	    return DAMSEL_DATA_TYPE_DOUBLE;
	default:
	    return NC_NAT;
    }
}

typedef enum {
        NC_UNSPECIFIED = 0, 
/* future       NC_BITFIELD = 7, */
/*      NC_STRING =     8,      */
        NC_DIMENSION =  10,
        NC_VARIABLE =   11,
        NC_ATTRIBUTE =  12,
        /* gross (these are not netcdf types), but helpful for damsel prototype */
        NC_DMSL_CONTAINER = 13,
        NC_DMSL_DATA = 14
} kind;


/* mode flags */
#define NC_NOWRITE      0       /* default is read only */
#define NC_WRITE        0x0001  /* read & write */

/* mode flags for ncmpi_create */
#define NC_CLOBBER      0
#define NC_NOCLOBBER    0x0004  /* Don't destroy existing file on create */

#define NC_GLOBAL (-1)

#define DIM_HANDLE_MASK 0x1
#define ATTR_HANDLE_MASK 0x2
#define VAR_HANDLE_MASK 0x4

int Dim_Handle_Count = 0;
int Attr_Handle_Count = 0;
int Var_Handle_Count = 0;
int Dmsl_Container_Count = 0;
int Data_Handle_Count = 0;

damsel_handle make_handle(kind k, unsigned int increment) {
    damsel_handle h;
    int *i;

    switch (k) {
	case NC_DIMENSION:
	    i = &Dim_Handle_Count;
	    break;
	case NC_ATTRIBUTE:
	    i = &Attr_Handle_Count;
	    break;
	case NC_VARIABLE:
	    i = &Var_Handle_Count;
	    break;
	case NC_DMSL_CONTAINER:
	    i = &Dmsl_Container_Count;
	    break;
	case NC_DMSL_DATA:
	    i = &Data_Handle_Count;
	    break;
	default:
	    i=NULL;
    }

    assert (i!= NULL);

    h =( (k<<16) | (*i) );
    (*i) = (*i) + increment;
    return h;
}

typedef struct NC {
    MPI_Comm comm;
    char *path;
    damsel_model dmsl_model;
    damsel_model dmsl_model_file;
} NC;

#define MAXDIMS 16
typedef struct NC_dim {
    char *name;
    MPI_Offset len;
    int dimid;
} NC_dim;

NC_dim DimArray[MAXDIMS];

/* length in bytes: we need this so we can pre-declare attribute types for both
 * reading and writing.  we need to pre-declare datatatypes for reading because
 * that's the present (early 2013) limitation of the damsel implementation */
#define MAX_ATTR_LEN 256    
#define NAME_MAX 256
typedef struct NC_attr {
    char name[NAME_MAX];
    int type;
    MPI_Offset len;
    int varid;      /* not needed in real pnetcdf */
    char data[MAX_ATTR_LEN];
} NC_attr;

#define MAXATTRS 8
NC_attr AttrArray[MAXATTRS];

#define MAXVARS 32
typedef struct NC_var {
    char *name;
    int type;
    int ndims;
    int dimarray[MAXDIMS];
    /* a pnetcdf variable needs to record
     * - name
     * - type
     * - dimension references
     * - data */
    damsel_data_type damsel_type;
    damsel_data mem_dim;
    damsel_data file_dim;
    damsel_data mem_data;
    damsel_data file_data;
    int varid;
} NC_var;

NC_var VarArray[MAXVARS];


typedef struct typedict {
    damsel_data_type attrtype;
    damsel_data_type dimtype;
    damsel_data_type vartype;
    int is_initialized;
} type_dictionary;

type_dictionary DamselTypes;

damsel_library Lib_Id; 


void make_damsel_types() {

    if (DamselTypes.is_initialized) return;

    size_t namelen=NAME_MAX, dimidlen=MAXDIMS, attrlen=MAX_ATTR_LEN;
    damsel_data_type nametype = DMSLtype_array_create(DAMSEL_DATA_TYPE_BYTES, 
	    1, &namelen);

    char *dimtype_labels[3] = {"name", "length", "dimid"};
    damsel_data_type dimtypes[3];
    dimtypes[0] = nametype;
    dimtypes[1] = DAMSEL_DATA_TYPE_INT64;
    dimtypes[2] = DAMSEL_DATA_TYPE_INTEGER;

    size_t dim_offsets[3];
    dim_offsets[0] = offsetof(NC_dim, name);
    dim_offsets[1] = offsetof(NC_dim, len);
    dim_offsets[2] = offsetof(NC_dim, dimid);

    DamselTypes.dimtype = DMSLtype_struct_create(3, dimtype_labels,
	    dimtypes, dim_offsets, sizeof(NC_dim));

    char *vartype_labels[5] = {"name",  "type", "ndims", "dimensions", "varid",};
    damsel_data_type vartypes[5];
    vartypes[0] = nametype;
    vartypes[1] = DAMSEL_DATA_TYPE_INTEGER;
    vartypes[2] = DAMSEL_DATA_TYPE_INTEGER;
    vartypes[3] = DMSLtype_array_create(DAMSEL_DATA_TYPE_INT64, 1, &dimidlen);
    vartypes[4] = DAMSEL_DATA_TYPE_INTEGER;

    size_t var_offsets[5];

    var_offsets[0] = offsetof(NC_var, name);
    var_offsets[1] = offsetof(NC_var, type);
    var_offsets[2] = offsetof(NC_var, ndims);
    var_offsets[3] = offsetof(NC_var, dimarray);
    var_offsets[4] = offsetof(NC_var, varid);

    DamselTypes.vartype = DMSLtype_struct_create(5, vartype_labels, vartypes, 
	    var_offsets, sizeof(NC_var));

#define NUM_ATTR_MEMBERS 5
    char *attrtype_lables[NUM_ATTR_MEMBERS] = {
	"name", "type", "length", "varid", "payload"};
    damsel_data_type attrtypes[NUM_ATTR_MEMBERS];
    attrtypes[0] = nametype; 
    attrtypes[1] = DAMSEL_DATA_TYPE_INTEGER; 
    attrtypes[2] = DAMSEL_DATA_TYPE_INT64;
    attrtypes[3] = DAMSEL_DATA_TYPE_INTEGER; 
    attrtypes[4] = DMSLtype_array_create(DAMSEL_DATA_TYPE_BYTES, 1, &attrlen);

    size_t attroffsets[NUM_ATTR_MEMBERS];
    attroffsets[0] = offsetof(NC_attr, name);
    attroffsets[1] = offsetof(NC_attr, type);
    attroffsets[2] = offsetof(NC_attr, len);
    attroffsets[3] = offsetof(NC_attr, varid);
    attroffsets[4] = offsetof(NC_attr, data);

    DamselTypes.attrtype = DMSLtype_struct_create(NUM_ATTR_MEMBERS, attrtype_lables, 
	    attrtypes, attroffsets, sizeof(NC_attr));

    DamselTypes.is_initialized=1;
    return;
}


int fake_ncmpi_create(MPI_Comm comm, char * path, int mode, NC *ncid) {
    damsel_err_t err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &ncid->dmsl_model);

    if ( ! (mode & NC_NOCLOBBER) ) unlink (path);

    damsel_trait_bundle tb;
    err = DMSLtrait_create(&tb);
    if ( (mode & NC_WRITE) )
	DMSLtrait_put(tb, "mode", "WRITE");
    else
	DMSLtrait_put(tb, "mode", "READ");

    DMSL_CHECK(DMSLmodel_attach(ncid->dmsl_model, path, comm, tb));

    DMSLtrait_release(tb);
    
    /* this is a yucky bit of API leakage from some earlier damsel ideas.   */
    damsel_model_internal *model_internal = (damsel_model_internal *)ncid->dmsl_model;
    ncid->dmsl_model_file = (damsel_model) model_internal->attached_file;

    /* if we had netcdf init/finalize functions we could put this type creation
     * bit there.. */

    make_damsel_types();

    ncid->comm = comm;
    ncid->path = strdup(path);
    return 0;
}

int fake_ncmpi_open(MPI_Comm comm, char * path, int mode, NC *ncid) {
    /* well, open and create are basically the same thing damsel-wise */
    fake_ncmpi_create(comm, path, mode, ncid);

    /* TODO: all the damsel stuff we do in fake_ncmpi_inq should go here */

    return 0;
}

int fake_ncmpi_close(NC *ncid) {
    DMSL_CHECK(DMSLmodel_close(ncid->dmsl_model));
    return 0;
}

int fake_ncmpi_enddef(NC *ncid) {
    
    /* maybe a bit too netcdf-specific, but at the end of define mode we'll
     * write out the variable, dimension, and attribute information */

    /* Damsel prototype can only handle a single transfer, so write md, close,
     * and reopen */

    DMSLmodel_transfer_sync(ncid->dmsl_model, DAMSEL_TRANSFER_TYPE_WRITE);
    DMSL_CHECK(DMSLmodel_close(ncid->dmsl_model));
    ncid->dmsl_model = DMSLmodel_attached_with_trait(ncid->comm, 
	    ncid->path, "REOPEN");

    return 0;
}

int fake_ncmpi_def_dim(NC *ncid, const char *name, MPI_Offset len, int *dim) 
{

    damsel_handle dim_tag_handle;
    damsel_handle dimlen_collection_handle;
    damsel_container dimlen_collection;
    damsel_err_t err;
    static int dimid=0;
    NC_dim *dimp = &(DimArray[dimid]);
    int rank;

    char newname[NAME_MAX];
    snprintf(newname, NAME_MAX, "dim-%d-%s", dimid, name);
    newname[NAME_MAX-1] = '\0';

    MPI_Comm_rank(ncid->comm, &rank);

    dim_tag_handle = make_handle(NC_DIMENSION, 1);

    DMSL_CHECK(DMSLtag_define(ncid->dmsl_model, &dim_tag_handle, DAMSEL_DATA_TYPE_INT64, newname));

    /* there isn't much to collect here: just one value (the dimension length): */

    dimlen_collection_handle = make_handle(NC_DMSL_CONTAINER, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model,
	    dimlen_collection_handle, 1, 1, &dimlen_collection);

    dimp->len = len;
    if (rank == 0) { /* only rank 0 "owns" the netcdf metadata, and is the only
			one that should map into the file */
	/* assign the size of the dimension to the tag representing our
	 * dimemsion */
	/* probably need a global mapping here to show that N processes all defined
	 * the same dimenson ? */
	/* let's try having only rank 0 map it */
	DMSL_CHECK(DMSLmodel_map_tag(&(dimp->len), dimlen_collection,
		    &dim_tag_handle));

	DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(
		    dimlen_collection));

	DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(
		    ncid->dmsl_model, dim_tag_handle));
    }

    dimp->dimid = dimid;

    *dim = dimid++;

    return 0;
}

int fake_ncmpi_put_att_text(NC *ncid, int varid, const char *name, MPI_Offset len, 
	const char *data)
{
    NC_attr *attr; /* the pnetcdf library sets this up for us, but need our own
		     in this standalone example */
    attr = calloc(1, sizeof(NC_attr));  /* cannot later-map stack-allocated data */
    int rank, nprocs;
    damsel_container attrval_container;
    damsel_err_t err;

    /* attributes are a little more complicated than dimensions: 
     * - they are associated with something (either a variable or the entire
     *   dataset)
     * - attributes have a type (dimensions are always MPI_Offset)
     * - attributes have variable length */

    damsel_handle attr_tag_handle;

    snprintf(attr->name, NAME_MAX, "%s", name);
    attr->name[NAME_MAX-1] = '\0';
    attr->type=NC_CHAR;
    attr->len=len;
    attr->varid=varid;
    memcpy(&(attr->data), data, len);

    MPI_Comm_rank(ncid->comm, &rank);
    MPI_Comm_size(ncid->comm, &nprocs);

    attr_tag_handle = make_handle(NC_ATTRIBUTE, 1);
    DMSL_CHECK(DMSLtag_define(ncid->dmsl_model, &attr_tag_handle, 
	    DamselTypes.attrtype, attr->name));

    err = DMSLcontainer_create_sequence(ncid->dmsl_model,
	    (damsel_handle)attr,  /* using memory address as handle */
	    1, 1, &attrval_container);

    if (rank == 0) {
	DMSL_CHECK(DMSLmodel_map_tag(attr, attrval_container, 
		    &attr_tag_handle));
	DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(
		    attrval_container));
	DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(
		    ncid->dmsl_model, attr_tag_handle));
    }

    return 0;
}

int fake_ncmpi_def_var(NC *ncid, const char *name, nc_type type, 
	int ndims, const int *dimidsp, int *varidp)
{
    /* netcdf defines variables by name, type, and the dimensions composing
     * that variable.  
     * e.g "a 1d floating point variable named temperature of length 10"
     * e.g. "a 3d integer variable named location, dimensions 10x20x30" 
     *
     * A struct type (see attributes) can do the trick, but the interface is
     * challenging to work with: there's no way to inspect types, and types for
     * reading must be the same as types created for writing.  So, encode the
     * stuff in the name 
     *
     * Main task: associate N dimension identifiers with a name */
    damsel_handle dimids_base_handle;
    damsel_err_t err;
    int rank;

    int i;

    static int varid=0;

    MPI_Comm_rank(ncid->comm, &rank);

    NC_var *varp = &(VarArray[varid]);
    varp->name = malloc(NAME_MAX);
    strncpy(varp->name, name, NAME_MAX);

    char newname[NAME_MAX];
    snprintf(newname, NAME_MAX, "var-%d-%s", varid, name);
    newname[NAME_MAX-1] = '\0';

    varp->mem_dim.tag_handle = make_handle(NC_VARIABLE, 1);
    varp->damsel_type = netcdf_type_to_damsel_type(type);
    if (varp->damsel_type == DAMSEL_DATA_TYPE_INVALID) {
	fprintf(stderr, "unsupported pnetcdf-on-damsel type\n");
	return -1;
    }
    DMSL_CHECK(DMSLtag_define(ncid->dmsl_model, &(varp->mem_dim.tag_handle),
	    varp->damsel_type, newname));

    dimids_base_handle=make_handle(NC_DIMENSION, ndims);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    dimids_base_handle, ndims, 1, &varp->mem_dim.container);

    varp->varid = varid;
    varp->type = type;
    varp->ndims = ndims;
    for (i=0; i<ndims; i++) 
	varp->dimarray[i] = dimidsp[i];


    if (rank == 0 ) {
	DMSL_CHECK(DMSLmodel_map_tag(varp->dimarray, varp->mem_dim.container, 
		&(varp->mem_dim.tag_handle) ));
	DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(
		    varp->mem_dim.container));
	DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(
		    ncid->dmsl_model, varp->mem_dim.tag_handle));
    }

    *varidp = varid++;

    /* this takes care of the dim (dimension) side of things.  We'll deal with
     * data in another step */

    return 0;
}

int fake_ncmpi_put_vara_int_all(NC *ncid, int varid, const MPI_Offset start[], 
	const MPI_Offset count[], const int *op)
{
    damsel_handle vardata_tag_handle;
    damsel_container vardata_container;
    damsel_err_t err;
    int nelems=1;
#ifdef USE_MAP_GLOBAL_HANDLES
    int rank;
#endif

    nelems = DMSL_subprod(VarArray[varid].ndims, (int64_t*)count);

    /* still a little stuck on generalizing to N-d..*/
    /* at this point in the pnetcdf library, though, we've got a nice tidy
     * buffer type converted and endian-swapped and ready to go */
    char newname[NAME_MAX];
    snprintf(newname, NAME_MAX, "vardata-%d-%s", varid, VarArray[varid].name);
    newname[NAME_MAX-1] = '\0';

    if (VarArray[varid].type != NC_INT) {
	fprintf(stderr, "unsupported pnetcdf-on-damsel type\n");
	return -1; /* bad type */
    }
    vardata_tag_handle = make_handle(NC_DMSL_DATA, 1);
    DMSL_CHECK(DMSLtag_define(ncid->dmsl_model, &vardata_tag_handle,
	    DAMSEL_DATA_TYPE_INTEGER, newname));

    /* each process takes process-local data and informs Damsel about it */
    /* in read case we call DMSLselect_subset_from_tag.  less clear if that's
     * effective in write case: select_subset presumes a bunch of (file side)
     * handles already exist */
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    (damsel_handle)op, /* using address as handle */
	    nelems,
	    sizeof(int), &vardata_container);
            /* TODO: figure out how to deal with n-dimensionally strided data */

    /* in a parallel netcdf write/read we have N processors operating on
     * different regions of a variable.  Every process has the same varid.   */

    DMSL_CHECK(DMSLmodel_map_tag((int *)op, vardata_container, &vardata_tag_handle));
    
#ifdef USE_MAP_GLOBAL_HANDLES
    /* now the fun part.  the non-rank-zero procs contribute their handle
     * mappings to rank 0 */

    /* this is, i think, how parallel I/O to a single damsel entity is supposed
     * to work.  Ben thinks that map_global_handles was intended for only small
     * amounts of data (?) */

    MPI_Comm_rank(ncid->comm, &rank);
    if (rank != 0) {
	damsel_container tag_container;
	int tagrank = 0;
	int *proc_ranks = calloc(nelems, sizeof(int)); /* all zeros: rank 0 has
                                                          the official mapping */
	DMSL_CHECK( DMSLmodel_map_global_handles(vardata_container, proc_ranks, 
		(damsel_handle*)op));
	err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
		vardata_tag_handle, 1, 1, &tag_container);
	DMSL_CHECK(DMSLmodel_map_global_handles(tag_container, &tagrank, &vardata_tag_handle));
    }

    if (rank == 0) {
        damsel_container full_container;
        damsel_handle vardata_collection_handle = make_handle(NC_VARIABLE, 1);
        /* rank 0 stores the mapping, but everyone writes data */
        DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(ncid->dmsl_model, 
		vardata_tag_handle));

        err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
		(damsel_handle)op, nelems, sizeof(int), &full_container);
        DMSL_CHECK(DMSLcoll_create(ncid->dmsl_model, &vardata_collection_handle, 
                full_container, DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR));
        DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(ncid->dmsl_model, 
                vardata_collection_handle));
        DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(full_container));
    }
#else 

    /* since map_global_handles won't do what I want, map the data one chunk
     * per processor */

    DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(vardata_container));
    DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(ncid->dmsl_model, vardata_tag_handle));
#endif


#if 1
    /* DMSL limitation: we can only perform one write transfer per open.  */
    DMSLmodel_transfer_sync(ncid->dmsl_model, DAMSEL_TRANSFER_TYPE_WRITE);
    DMSL_CHECK(DMSLmodel_close(ncid->dmsl_model));
    ncid->dmsl_model = DMSLmodel_attached_with_trait(ncid->comm, ncid->path, "REOPEN");
#endif

    return NC_NOERR;
}
int fake_ncmpi_get_vara_int_all(NC *ncid, int varid, const MPI_Offset start[],
	const MPI_Offset count[], int *op) 
{
    int64_t nitems, *size;
    NC_var var;
    char vardataname[NAME_MAX];
    int i;
    damsel_err_t err;

    /* model gets closed and reopened after every read, so need to repopulate */
    DMSL_CHECK(DMSLmodel_populate(ncid->dmsl_model));

    var = VarArray[varid];

    size = calloc(var.ndims, sizeof(int64_t));
    /* remember: the dimarray is an array of identifiers, not sizes */
    for(i=0; i< var.ndims; i++ )
	size[i] = DimArray[var.dimarray[i]].len;

    /* in inquiry, we obtained the dimension identifiers (the shape).  here we
     * use that and the reqeusted start and count to actually do I/O */

    snprintf(vardataname, NAME_MAX, "vardata-%d-%s", varid, VarArray[varid].name);
    err = DMSLselect_tag_by_name(ncid->dmsl_model, vardataname, &var.file_data.tag_handle);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model_file, 
	    var.file_data.tag_handle, 1, 1, &var.file_data.tag_container);

    var.mem_data.tag_handle = make_handle(NC_VARIABLE, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    var.mem_data.tag_handle, 1, 1, &var.mem_data.tag_container);

    DMSL_CHECK(DMSLmodel_map_handles(var.mem_data.tag_container,
		var.file_data.tag_container));

    assert (sizeof (MPI_Offset) == sizeof (int64_t));

    nitems = DMSL_subprod(var.ndims, (int64_t*)count);

    var.file_data.container = DMSLselect_subset_from_tag(ncid->dmsl_model,
	    var.file_data.tag_handle, var.ndims, size, (int64_t *)
	    start, (int64_t*) count);

    var.mem_data.container_id = make_handle(NC_DMSL_CONTAINER, nitems);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    var.mem_data.container_id, nitems, 1, &var.mem_data.container);

    DMSL_CHECK(DMSLmodel_map_handles(var.mem_data.container,
		var.file_data.container)); 
    DMSL_CHECK(DMSLmodel_map_tag(op, var.mem_data.container, 
		&(var.mem_data.tag_handle)));
    
    /* while multiple read transfers *work*, that's more out of good fortune
     * than by design.  */
    DMSLmodel_transfer_sync(ncid->dmsl_model, DAMSEL_TRANSFER_TYPE_READ);
    DMSL_CHECK(DMSLmodel_close(ncid->dmsl_model));
    ncid->dmsl_model = DMSLmodel_attached_with_trait(ncid->comm, ncid->path, "REOPEN");

    return NC_NOERR;

}

void decode_vartag(NC *ncid, damsel_tag_buf_type tag)
{
    int varid; 
    char varname[NAME_MAX];
    NC_var *var;
    damsel_err_t err;

    /* names look like var-varid-name; */
    /* improperly-formatted damsel input files could trivally overrun these
     * buffers */
    sscanf(tag.name, "var-%d-%s", &varid, varname);
    var = &(VarArray[varid]);
    
    var->name = strdup(varname);
    var->varid = varid;
    var->type = damsel_type_to_netcdf_type(tag.tag_datatype);
    /* dimarray_tag_handle is  memory-side and therefore ephermeral */
    var->mem_dim.tag_handle = make_handle(NC_DMSL_CONTAINER, 1);  

    /* before defining a (sequence) handle container, need to find out how big
     * container is in file */

    /* this is just the tag in file, not the contianer of dimids we need.  i actually
     * have the tag handle in the tag structure (caller of this function called
     * get_tuple_count/get_tag_info), so this call is redundant */
    err =  DMSLselect_tag_by_name(ncid->dmsl_model, tag.name, &var->file_dim.tag_handle);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model_file,
	    var->file_dim.tag_handle, 1, 1, &var->file_dim.tag_container);

    /* set up an application-side tag for the application-side container */
    err = DMSLcontainer_create_sequence(ncid->dmsl_model,
	    var->mem_dim.tag_handle, 1, 1, &var->mem_dim.tag_container);

    /* note: not mapping any data; just the "buckets" to set up a later transfer */
    DMSL_CHECK(DMSLmodel_map_handles( var->mem_dim.tag_container, 
		var->file_dim.tag_container));

     /* how big is the collection associated with this tag?  before we
     * 'container_count', we need a container.. */
    var->file_dim.container = DMSLselect_handles_with_values(ncid->dmsl_model, 
	    var->file_dim.tag_handle);

    var->ndims = DMSLcontainer_count(var->file_dim.container);
    assert(var->ndims < MAXDIMS);

    err = DMSLcontainer_create_sequence(ncid->dmsl_model,
	    var->mem_dim.container_id, var->ndims, 1, &var->mem_dim.container);

    /* associate tag with container-of-memory */
    DMSL_CHECK(DMSLmodel_map_tag(var->dimarray, var->mem_dim.container, 
		&(var->mem_dim.tag_handle)));

    /* we right now don't actually need the values of the dimension
     * identifiers.  It's up to our caller to initiate the actual transfer. */
    /* said differently, the 'dimarray' holds the shape of a big blob of data.
     * To read a variable we need both the shape and the blob */

    /* finally, map values from container in file to container in memory */
    DMSL_CHECK(DMSLmodel_map_handles(var->mem_dim.container, var->file_dim.container))

}

void decode_dimtag(NC *ncid, damsel_tag_buf_type tag)
{
    int dimid;
    char dimname[NAME_MAX];
    damsel_data memory, file;
    damsel_err_t err;

    sscanf(tag.name, "dim-%d-%s", &dimid, dimname);
    DimArray[dimid].name = strdup(dimname);
    DimArray[dimid].dimid = dimid;


    /* dimensions have just a single value associated with them:  we will read
     * from the file into DimArray[dimid].len */
    /* set up memory-side buckets*/
    memory.tag_handle = make_handle(NC_DIMENSION, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    memory.tag_handle, 1, 1, &memory.tag_container);
    memory.container_id = make_handle(NC_DMSL_DATA, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    memory.container_id, 1, 1, &memory.container);

    /* we actually have this already but for clarity, turn name into fileside container */
    err = DMSLselect_tag_by_name(ncid->dmsl_model, tag.name, &file.tag_handle);
    err =DMSLcontainer_create_sequence(ncid->dmsl_model_file,
	    file.tag_handle, 1, 1, &file.tag_container);
    /* map file-side tag-handles into memory */
    DMSL_CHECK(DMSLmodel_map_handles (memory.tag_container, file.tag_container));


    DMSL_CHECK(DMSLmodel_map_tag(&(DimArray[dimid].len), memory.container, 
		&(memory.tag_handle)));

    file.container = DMSLselect_handles_with_values(ncid->dmsl_model,
	    file.tag_handle);

    DMSL_CHECK(DMSLmodel_map_handles(memory.container, file.container));

    /* here again, we'll make the transfer later, as that's the rule for damsel now */
}

void decode_attrtag(NC *ncid, damsel_tag_buf_type tag)
{
    char attrname[NAME_MAX];
    NC_attr attr;   /* note: this stack-allocated thing will have get copied
		       back to the global attribute dictionary.  This
		       dictionary is indexed based on 'varid', and we don't
		       know the varid until we read everything.   */
    damsel_data memory, file;
    damsel_err_t err;

    sscanf(tag.name, "attr-%s", attrname);

    /* attribute information is kind of like a dimension array: could be
     * variable length */
    err = DMSLselect_tag_by_name(ncid->dmsl_model, tag.name, &file.tag_handle);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model_file, 
	    file.tag_handle, 1, 1, &file.tag_container);

    memory.tag_handle = make_handle(NC_DMSL_CONTAINER, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model,
	    memory.tag_handle, 1, 1, &memory.tag_container);

    DMSL_CHECK(DMSLmodel_map_handles( memory.tag_container, file.tag_container));

    file.container = DMSLselect_handles_with_values(ncid->dmsl_model, 
	    file.tag_handle);
    /* DMSLcontainer_count will report the number of values this tag has
     * assigned, but if we use struct types, the count is just "one type" */

    memory.container_id = make_handle(NC_DMSL_CONTAINER, 1);
    err = DMSLcontainer_create_sequence(ncid->dmsl_model, 
	    memory.container_id, 1, 1, &memory.container);

    /* finally, map values to the area in memory */
    DMSL_CHECK(DMSLmodel_map_handles(memory.container, file.container));
    DMSL_CHECK(DMSLmodel_map_tag(&attr, memory.container, &(memory.tag_handle)));

    DMSLmodel_transfer_sync(ncid->dmsl_model, DAMSEL_TRANSFER_TYPE_READ);

    /* now copy from stack into more permanently available memory */
    if (attr.varid == NC_GLOBAL) {
	memcpy(&(AttrArray[MAXATTRS-1]), &attr, sizeof(attr));
    } else {
	fprintf(stderr, "UH, don't know how to handle non-global attributes\n");
    }
}

int fake_ncmpi_inq(NC *ncid, int* ndims, int *nvars, int *natts, int * isrecvar) {

    int container_cnt, tag_cnt, i;
    damsel_tag_buf_type *all_tags;

    damsel_request_t req;
    damsel_status_t status;

    /* no longer populate at open. populate on-demand .  Well, besides we want to move most of this stuff into ncmpi_open anyway */
    DMSL_CHECK(DMSLmodel_populate(ncid->dmsl_model));

    /* what tools do we have for inquiry?  DMSLselect_tag_by_name, well, needs
     * the name.   might be fun to overload what 'entity' means: quad ->
     * attribute; poly -> dimension; prism ->variable.*/
     DMSLmodel_get_tuple_count(ncid->dmsl_model, &container_cnt, &tag_cnt);

     all_tags = calloc(tag_cnt, sizeof(damsel_tag_buf_type));
     DMSL_CHECK(DMSLmodel_get_tag_infos(ncid->dmsl_model,  all_tags));

     /* TODO: need to store metadata as a type, but data goes into a separate
      * tag. that way we can discrimiate on damsel data types, not on encoding
      * anything into the name */
     *ndims = *nvars = *natts = 0;
     for(i=0; i<tag_cnt; i++) {
	 if (strncmp(all_tags[i].name, "var-", strlen("var-") ) == 0) {
	     decode_vartag(ncid, all_tags[i]);
	     (*nvars)++;
	 } else if (strncmp(all_tags[i].name, "dim-", strlen("dim-") ) ==0) {
	     decode_dimtag(ncid, all_tags[i]);
	     (*ndims)++;
	 } else if (all_tags[i].tag_datatype == DamselTypes.attrtype) {
	     decode_attrtag(ncid, all_tags[i]);
	     (*natts)++;
	 }
     }

    DMSL_CHECK(DMSLmodel_transfer_async(ncid->dmsl_model, DAMSEL_TRANSFER_TYPE_READ, &req));
    DMSL_CHECK(DMSLmodel_wait(req, &status));
    
    *isrecvar = 0; /* pretend we have no record variables */

    return 0;
}

int fake_ncmpi_inq_dimlen(NC *ncid, int dimid, int *size) {
    *size = DimArray[dimid].len;
    return 0;
}

int fake_ncmpi_inq_var(NC *ncid, int varid, char *name, nc_type *type, 
	int *ndims, int *dimids, int *natts) 
{
    int i;
    /* need some logic in here to do tag processing if ncmpi_inq was not
     * already called. Maybe we do the tag processing when we open the file? */
    strcpy(name, VarArray[varid].name);
    /* type is in the vardata array */
    *type = VarArray[varid].type;
    /* ndims and dimids are in the var sequence */
    *ndims = VarArray[varid].ndims;
    for (i=0; i<*ndims; i++) dimids[i] = VarArray[varid].dimarray[i];
    /* TODO: where did we store the number of associated attributes? */
    *natts = 0;  /* BUG: there are attributes ... somewhere */
    return 0;
}


#define DATA_PER_PROC 12

int test_pnetcdf_write(MPI_Comm comm, char *name)  {

    int nprocs, rank;
    char buf[13] = "Hello World\n";
    int dimarray[3];

    MPI_Comm_size(comm, &nprocs);
    MPI_Comm_rank(comm, &rank);


    NC ncid;
    int dim1, dim2, dim3, dim4, dim5;
    int v1, v2;
    MPI_Offset start[3], count[3];
    int *data, *data2;
    int i;

    data = calloc(DATA_PER_PROC, sizeof(int));

    fake_ncmpi_create(comm, name,  NC_WRITE|NC_CLOBBER, &ncid);

    fake_ncmpi_def_dim(&ncid, "d1", nprocs*DATA_PER_PROC, &dim1);
    fake_ncmpi_def_dim(&ncid, "d2", nprocs*2, &dim2);
    fake_ncmpi_def_dim(&ncid, "d3", nprocs*4, &dim3);
    fake_ncmpi_def_dim(&ncid, "d4", DATA_PER_PROC, &dim4);
    fake_ncmpi_def_dim(&ncid, "d5", nprocs, &dim5);
    
    fake_ncmpi_put_att_text(&ncid, NC_GLOBAL, "string", 13, buf);

    /* simple test of 1d variable */
    dimarray[0] = dim1;
    fake_ncmpi_def_var(&ncid, "v1", NC_INT, 1, dimarray, &v1);

    fake_ncmpi_enddef(&ncid);

    /* simple 1d case */
    start[0] = rank; count[0]=DATA_PER_PROC;
    for (i=0; i<DATA_PER_PROC; i++) data[i]=(1000+(rank*100)+i);

    fake_ncmpi_put_vara_int_all(&ncid, v1, start, count, data);

    /* a simple 2d var */
    data2 = calloc(DATA_PER_PROC, sizeof(int));
    for (i=0; i<DATA_PER_PROC; i++) data2[i]=(1000+(rank*100)+i);
    dimarray[0] = dim5;
    dimarray[1] = dim4;
    fake_ncmpi_def_var(&ncid, "v2d", NC_INT, 2, dimarray, &v2);

    start[0] = rank; start[1] = 0;
    count[0] = 1;    count[1] = DATA_PER_PROC;
    fake_ncmpi_put_vara_int_all(&ncid, v2, start, count, data2);
    
    fake_ncmpi_close(&ncid);

    return 0;
}

void dump_var_int(char *name, int size, int *buf) 
{
    int i;
    fprintf(stderr, "varname: %s\n", name);
    for (i=0; i<size; i++) 
	fprintf(stderr, "||%d", buf[i]);
    fprintf(stderr, "||\n");
}


int test_pnetcdf_read(MPI_Comm comm, char *name) {
    NC ncid;
    int i, j, ndims, nvars, natts, isrecvar, var_ndims, var_natts;
    int nprocs, rank;
    int *dim_sizes, var_size;
    int dimids[MAXDIMS];
    char varname[NAME_MAX];
    MPI_Offset *start, *count;
    nc_type type;
    int *data, ret;

   
    /* the streamlined "know nothing" approach demonstrated in pnetcdf's
     * examples/tutorial/pnetcdf-read-standard.c -- we pretend we did not just
     * write the file and imagine that we do not know what's in the file.  
     * - Open the file, 
     * - enquire how many "things" are in the file
     * - allocate enough space
     * - get dimension sizes
     * - get variable name, type, and shape 
     * - finally, actually read the variable */
    
    fake_ncmpi_open(comm, name, NC_NOWRITE|NC_NOCLOBBER, &ncid);
    fake_ncmpi_inq(&ncid, &ndims, &nvars, &natts, &isrecvar);

    fprintf(stderr, "=== attrval: %s == \n", (char *)(AttrArray[MAXATTRS-1].data));

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &nprocs);

    dim_sizes = calloc(ndims, sizeof(int));
    
    for (i=0; i<ndims; i++) {
	fake_ncmpi_inq_dimlen(&ncid, i, &(dim_sizes[i]));
    }

    for (i=0; i<nvars; i++ ) {
	fake_ncmpi_inq_var(&ncid, i, varname, &type, &var_ndims, 
		dimids, &var_natts);
	start = calloc(var_ndims, sizeof(MPI_Offset));
	count = calloc(var_ndims, sizeof(MPI_Offset));

	/* decomposistion depends on number of dimensions */
	switch (var_ndims) {
	    case 1:
		start [0] = (dim_sizes[dimids[0]]/nprocs)*rank;
		count[0] = (dim_sizes[dimids[0]]/nprocs);
		break;
	    case 2:
		start[0] = rank; start[1] = 0;
		count[0] = 1   ; count[1] = DATA_PER_PROC;
		break;
	    default:
		fprintf(stderr, "Uhoh: do not know how to decompose %d dims\n", 
			var_ndims);
	}

	var_size = count[0];
	for (j=1; j<var_ndims; j++) {
	    var_size *= count[j];
	}
	switch (type) {
	    case NC_INT:
		data = calloc(var_size, sizeof(int));
		ret = fake_ncmpi_get_vara_int_all(&ncid, i, start, count, data);
		assert(ret == NC_NOERR);
		dump_var_int(varname, var_size, data);
		break;
	    default:
		fprintf(stderr, "unsupported NetCDF type\n");
	}
	free (start);
	free (count);
    }
    if (data != NULL) free (data);
    return 0;
}

int main(int argc, char ** argv)
{
    /* there's no "pnetcdf init", so we do damsel init here */
    int lvl;
#ifndef DAMSEL_SINGLETHREAD
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &lvl);
    assert(lvl == MPI_THREAD_MULTIPLE);
#else
    MPI_Init(&argc, &argv);
#endif

    damsel_library lib;
    damsel_err_t err = DMSLlib_init(&lib);

    test_pnetcdf_write(MPI_COMM_WORLD, argv[1]);

    test_pnetcdf_read(MPI_COMM_WORLD, argv[1]);

    /* there's no "pnetcdf finalize" so we do finialize stuff here */
    DMSLlib_finalize(lib);
    MPI_Finalize();

    return 0;
}
