/* tests reading into a sequence */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <unistd.h>
#include <memory.h>

#include "damsel.h"
#include "damsel-internal.h"

int main(int argc, char **argv) {

  // we'll read the data into data[] and then check that
  // it matches up with target data check that
  // it matches up with target data
  short int data[5];

  printf("%s\n", argv[0]);
  if(argc!=2) {
    printf("specify filename of .h5 dump file to check\n");
    return 2;
  }

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);
  damsel_model_internal *ms = (damsel_model_internal *)model;
  assert(ms != NULL);

  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");
  DMSLmodel_attach(model, argv[1], MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  DMSLmodel_populate(model);

  /* define a tag which will store our data. the tag will be
     named 'somedata'. this handle doesn't need to match up
     with the handle used to write data. */

  damsel_handle data_tag_handle = 0x31000;
  // this tag should appear model-side by virtue of mapping later on, rather
  // than because of an explicit tag_define - that is, the tag define we
  // called model-side at write time should be ending up in the model here
  // via a series of maps.
  // DMSLtag_define(model, &data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "somedata");


  /* define a collection of the handles on which we will store data. we
     will store the data on handles 20000 .. upwards and identify the
     collection of these handles by another handle, 10020. */

  /* we'll create a sequence handle collection with 5 handles in it. */

  damsel_container vertex_collection_idA;
  err = DMSLcontainer_create_sequence(model, 0x80900, 1, 0x5, &vertex_collection_idA);
  damsel_container vertex_collection_idB;
  err = DMSLcontainer_create_sequence(model, 0x80905, 4, 0x5, &vertex_collection_idB);

  assert(vertex_collection_idA != NULL);
  assert(vertex_collection_idB != NULL);

  damsel_container tag_collection_id;
  err = DMSLcontainer_create_sequence(model,
    data_tag_handle, 1, 1, &tag_collection_id);

  assert(tag_collection_id != NULL);

  damsel_handle fileside_data_tag_handle;
  err = DMSLselect_tag_by_name(model, "somedata", &fileside_data_tag_handle);
  damsel_container file_tag_collection_id;
  err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, fileside_data_tag_handle, 1, 1, &file_tag_collection_id);
  DMSLmodel_map_handles(tag_collection_id, file_tag_collection_id);

// these two containers need to be generated without using explicit handles.

// get container with all values
  damsel_container filestore_vertex_collection_id = DMSLselect_handles_with_values(model, fileside_data_tag_handle);

  damsel_handle vA;
  damsel_handle vB[4];

// first value using damsel_handle handle_in_container_struct(damsel_handle_container *hc, int i);

  vA = DMSLcontainer_handle_at_position(filestore_vertex_collection_id, 0);

  int ix;
  for(ix=0;ix<4;ix++) {
    vB[ix] = DMSLcontainer_handle_at_position(filestore_vertex_collection_id, ix+1);
  }
  DMSLcontainer_release(filestore_vertex_collection_id);

  damsel_container filestore_vertex_collection_idA;
  err = DMSLcontainer_create_vector((damsel_model)ms->attached_file, &vA, 1, &filestore_vertex_collection_idA);
  damsel_container filestore_vertex_collection_idB;
  err = DMSLcontainer_create_vector((damsel_model)ms->attached_file, vB, 4, &filestore_vertex_collection_idB);

  DMSLmodel_map_handles(vertex_collection_idA, filestore_vertex_collection_idA);
  DMSLmodel_map_handles(vertex_collection_idB, filestore_vertex_collection_idB);

  // so now I can look for the tag I want, by name, and that will give me its file-side handle.

  // then I should be able to map it locally, and that should cause
  // the memory-side tag definition to appear at execute time.

  damsel_request_t req;
  damsel_status_t status;

  data[1] = 9;


  DMSLmodel_start_read(model);
  DMSLmodel_map_tag(data, vertex_collection_idA, &data_tag_handle);

  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);
  DMSLmodel_wait(req, &status);
  DMSLmodel_stop_read(model);

  assert(data[0] == 3);
  data[0]=8;
  assert(data[1] == 9); // check that the above read didn't mutate data[1]

  DMSLmodel_map_tag(&(data[1]), vertex_collection_idB, &data_tag_handle);

  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);
  DMSLmodel_wait(req, &status);

  // here check that the first element is 5 so that we can see that the
  // first read didn't happen again
  assert(data[0] == 8);
  assert(data[1] == 1);
  assert(data[2] == 4);
  assert(data[3] == 1);
  assert(data[4] == 5);

  DMSLmodel_close(model);
  DMSLcontainer_close(vertex_collection_idA);
  DMSLcontainer_close(vertex_collection_idB);
  DMSLcontainer_close(file_tag_collection_id);
  DMSLcontainer_close(tag_collection_id);
  DMSLcontainer_close(filestore_vertex_collection_idA);
  DMSLcontainer_close(filestore_vertex_collection_idB);


  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

