/* test-subarray-read: first, blat out a big hunk of 3d data so we have
 * something to read. next, attempt to read that data in parallel. */

#include <stdio.h>
#include <unistd.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-sugar.h"

#include <mpi.h>

#include <assert.h>

#define FILENAME "output/test-subarray-read-parallel.dmsl"
#define TAGNAME "array"

//#define VERBOSE


void seed_buffer(int *buf, int ndims, int *dims)
{
    int i, j, k;
    /* i think there's a way to do this initialization recursively for any N
     * dimensions but I haven't worked it out yet */
    assert(ndims == 3);
    for (i=0; i< dims[0]; i++)
	for (j=0; j<dims[1]; j++)
	    for (k=0; k<dims[2]; k++)
		buf[i*dims[1]*dims[2] + j*dims[2] + k] = 
		    100*i*dims[1]*dims[2] + 10*j*dims[2] + k;
}

void seed_file(MPI_Comm comm, char *fname, int ndims, int *dims)
{
    damsel_model model;
    damsel_data stuffing;  /* happy thanksgiving */
    damsel_err_t err;
    int handle_counter = 900;
    int i,nitems;
    int *data;

    unlink(fname);
    model = DMSLmodel_attached_with_trait(comm, fname, "WRITE");

    nitems=1;
    for (i=0; i<ndims; i++) {
	if (dims[i] == 0) dims[i] = 1; /* what does a zero dimension mean? */
	nitems *= dims[i];
    }

    data = calloc(nitems, sizeof(int));
    seed_buffer(data, ndims, dims);

    err = DMSLcontainer_create_sequence(model, handle_counter++, 
	    nitems, 1, &stuffing.container);
    stuffing.tag_handle = handle_counter+=nitems;

    DMSL_CHECK(DMSLtag_define(model, &(stuffing.tag_handle), DAMSEL_DATA_TYPE_INTEGER,
	    TAGNAME));
    DMSL_CHECK(DMSLmodel_map_tag(data, stuffing.container, &(stuffing.tag_handle)));

    DMSL_CHECK(DMSLmodel_map_handles_inventing_file_handles(stuffing.container));
    DMSL_CHECK(DMSLmodel_map_single_handle_inventing_file_handle(model,
		stuffing.tag_handle));

    DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_WRITE);

    DMSL_CHECK(DMSLcontainer_release(stuffing.container));
    DMSL_CHECK(DMSLmodel_close(model));

    free(data);

    return;
}

int subarray_read_3d_parallel(MPI_Comm comm, char *filename) 
{
    /* 3d (and general Nd case) builds upon 2d */

    damsel_model model, fileside_model;
    damsel_model_internal *model_internal;
    damsel_data mem, file;
    damsel_container subset_container;
    damsel_err_t err;
    int nitems, *data, i;
    int ndims=3; 
    int64_t size[3], start[3], count[3];
    int nerrors=0;
    int rank, nprocs;
    int standard[4][30] = { { 0, 1, 2, 60, 61, 62, 
	120, 121, 122, 180, 181, 182, 
	240, 241, 242, 3000, 3001, 3002, 
	3060, 3061, 3062, 3120, 3121, 3122, 
	3180, 3181, 3182, 3240, 3241, 3242, },
	{ 6000, 6001, 6002, 6060, 6061, 6062, 6120, 6121, 6122, 
	    6180, 6181, 6182, 6240, 6241, 6242, 9000, 9001, 9002, 
	    9060, 9061, 9062, 9120, 9121, 9122, 
	    9180, 9181, 9182, 9240, 9241, 9242 },
	{3, 4, 5, 63, 64, 65, 123, 124, 125, 183, 184, 185, 
	    243, 244, 245, 3003, 3004, 3005, 
	    3063, 3064, 3065, 3123, 3124, 3125, 
	    3183, 3184, 3185, 3243, 3244, 3245},
	{ 6003, 6004, 6005, 6063, 6064, 6065, 6123, 6124, 6125, 
	    6183, 6184, 6185, 6243, 6244, 6245, 9003, 9004, 9005, 
	    9063, 9064, 9065, 9123, 9124, 9125, 
	    9183, 9184, 9185, 9243, 9244, 9245 } };

    int handle_counter = 100;

    /* slab-of-data lives in a notionally 4x5x6 array. most aggressive
     * parallelism would be 2x5x3 sub-cubes */

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &nprocs);

    model = DMSLmodel_attached_with_trait(comm, filename, "READ");
    DMSL_CHECK(DMSLmodel_populate(model));

    model_internal= (damsel_model_internal*)model;
    fileside_model = (damsel_model) model_internal->attached_file;

    err = DMSLselect_tag_by_name(model, TAGNAME, &file.tag_handle);
    err = DMSLcontainer_create_sequence(fileside_model, 
	    file.tag_handle,  1, 1, &file.tag_container);

    mem.tag_handle = handle_counter ++;
    err = DMSLcontainer_create_sequence(model,
	    mem.tag_handle, 1, 1, &mem.tag_container);

    DMSL_CHECK(DMSLmodel_map_handles(mem.tag_container, file.tag_container));

    size[0] = 4; size[1] = 5; size[2] = 6; 

    /* ugh. N-d decomposistion.  */
    count[0] = 2; count[1] = 5; count[2] = 3;
    switch(rank) {
	case 0:
	    start[0] = 0; start[1] = 0; start[2] = 0;
	    break;
	case 1:
	    start[0] = 2; start[1] = 0; start[2] = 0;
	    break;
	case 2:
	    start[0] = 0; start[1] = 0; start[2] = 3;
	    break;
	case 3:
	    start[0] = 2; start[1] = 0; start[2] = 3;
	    break;
	default:
	    start[0] = start[1] = start[2] = 0;
	    count[0] = count[1] = count[2] = 0;
    }

    nitems = DMSL_subprod(3, count);
    data = calloc(nitems, sizeof(int));

    subset_container = DMSLselect_subset_from_tag(model, file.tag_handle, 
	    ndims, size, start, count);
   

    mem.container_id = handle_counter; handle_counter += nitems;
    err = DMSLcontainer_create_sequence(model, mem.container_id,
	    nitems, 1, &mem.container);

    DMSL_CHECK(DMSLmodel_map_handles(mem.container, subset_container));
    DMSL_CHECK(DMSLmodel_map_tag(data, mem.container, &(mem.tag_handle)));

    DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_READ);

#ifdef VERBOSE
    file.container = DMSLselect_handles_with_values(model, file.tag_handle);
    int all = DMSLcontainer_count(file.container);
    fprintf(stderr, "read %d of %d: ", nitems, all);
    for (i=0; i<nitems; i++) {
	fprintf(stderr, "||%d", data[i]);
    }
    fprintf(stderr, "||\n");
    DMSL_CHECK(DMSLcontainer_release(file.container));
#endif
    for (i=0; i<nitems; i++) {
	if (standard[rank][i] != data[i])  {
	    nerrors++;
	    fprintf(stderr, "%d of %d: Expected %d got %d\n", rank, nprocs, standard[rank][i], data[i]);
	}
    }

    free(data);
    DMSL_CHECK(DMSLcontainer_release(mem.tag_container));
    DMSL_CHECK(DMSLcontainer_release(mem.container));
    DMSL_CHECK(DMSLcontainer_release(file.tag_container));
    DMSL_CHECK(DMSLcontainer_release(subset_container));
    DMSL_CHECK(DMSLmodel_close(model));
    return nerrors;
}
int main(int argc, char **argv) 
{
    int threadlvl, nprocs, rank;
#define NDIMS 3
    int dims[NDIMS] = {4, 5, 6};
    int nerrors=0, all_errors=0;;
    
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &threadlvl);
    assert(threadlvl == MPI_THREAD_MULTIPLE);
    damsel_library lib;
    damsel_err_t err = DMSLlib_init(&lib);

    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0) 
	seed_file(MPI_COMM_SELF, FILENAME, 3, dims);

    nerrors += subarray_read_3d_parallel(MPI_COMM_WORLD, FILENAME);

    MPI_Allreduce (&nerrors, &all_errors, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

    DMSLlib_finalize(lib);
    MPI_Finalize();
    if (all_errors == 0)
	return 0;
    else
	return -nerrors;
}
