/* tests reading into a sequence of array tags */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <unistd.h>
#include <memory.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-debug.h"

void free_filetags_container(void *m) {
  printf("freeing handle container (as void type) %p", m);
  damsel_container_internal *c = (damsel_container_internal *)m;
  printf("freeing handle container (as container type) %p", c);
  assert_valid_handle_container_id(c);
  DMSLcontainer_close((damsel_container) c);
}

int main(int argc, char **argv) {

  // we'll read the data into data[] and then check that
  // it matches up with target data check that
  // it matches up with target data
  short int targetdata[][3] = {{3, 3, 3}, {1, 1, 1}, {4, 4, 4}, {1, 1, 1}, {5, 5, 5}}; // some arbitrary data to write
  short int data[5][3];

  printf("%s\n", argv[0]);
  if(argc!=2) {
    printf("specify filename of .h5 dump file to check\n");
    return 2;
  }

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);
  damsel_model_internal *ms = (damsel_model_internal *)model;
  assert(ms != NULL);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  DMSLmodel_attach(model, argv[1], MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  DMSLmodel_populate(model);

  /* Create an array datatype to use for the data */
  size_t array_dim[1] = {3};
  damsel_data_type array_type = DMSLtype_array_create(DAMSEL_DATA_TYPE_SHORTINT, 1, array_dim);

  /* define a tag which will store our data. the tag will be
     named 'somedata'. this handle doesn't need to match up
     with the handle used to write data. */

  damsel_handle data_tag_handle = 0x31000;
  // Note: this tag definition is very "fragile" right now - it depends
  // on the damsel data types being defined in exactly the same way and exactly
  // the same order for the writer and reader.  In the future, we'll have to
  // do an actual comparison of the underlying datatypes. - QAK
  DMSLtag_define(model, &data_tag_handle, array_type, "somedata");


  /* define a collection of the handles on which we will store data. we
     will store the data on handles 20000 .. upwards and identify the
     collection of these handles by another handle, 10020. */

  /* we'll create a sequence handle collection with 5 handles in it. */

  damsel_container vertex_collection_id;
  err = DMSLcontainer_create_sequence(model, 0x80900, 5, 0x5, &vertex_collection_id);

  assert(vertex_collection_id != NULL);

  damsel_container tag_collection_id;
  err = DMSLcontainer_create_sequence(model, 
    data_tag_handle, 1, 1, &tag_collection_id);

  assert(tag_collection_id != NULL);

  int tag_infos_count;
  int found=0;
  assert(ms->attached_file != NULL);
  DMSLmodel_get_tuple_count(model, NULL, &tag_infos_count);
  size_t amt = tag_infos_count * sizeof(damsel_tag_buf_type);
printf("will malloc %zd bytes for tag buffers. sizeof buf = %zd. tag_infos_count = %zd\n", amt, sizeof(damsel_tag_buf_type), (size_t) tag_infos_count);
  damsel_tag_buf_type *tag_buf = malloc(amt);
  assert(tag_buf != NULL);
printf("getting tag infos\n");
  DMSLmodel_get_tag_infos(model, tag_buf);
printf("got tag infos\n");
  linkedlist *filetags_ll = NULL;
  int tag_ix;
  for(tag_ix=0; tag_ix < tag_infos_count; tag_ix++) {
printf("checking tag %d with name %s: ", tag_ix, tag_buf[tag_ix].name);
    if(!strcmp(tag_buf[tag_ix].name, "somedata")) {
printf("  matched.\n");
      assert(found == 0); // not found already
printf("  tag handle in file is 0x%llx\n", *tag_buf[tag_ix].tag_handle);
      damsel_container file_tag_collection_id;
      err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, *tag_buf[tag_ix].tag_handle, 1, 1, &file_tag_collection_id);
      DMSLmodel_map_handles(tag_collection_id, file_tag_collection_id);
      assert_valid_handle_container_id(file_tag_collection_id);
      linkedlist_prepend(&filetags_ll, file_tag_collection_id);
      found=1;
    } else {
printf("skipping.\n");
    }
  }
  free(tag_buf);

  assert(found == 1); // check we did find it somewhere.

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_collection_id, &data_tag_handle);

  damsel_handle fileside_data_tag_handle;
  err = DMSLselect_tag_by_name(model, "somedata", &fileside_data_tag_handle);

  damsel_container filestore_vertex_collection_id = DMSLselect_handles_with_values(model, fileside_data_tag_handle);

  DMSLmodel_map_handles(vertex_collection_id, filestore_vertex_collection_id);
  DMSLcontainer_release(filestore_vertex_collection_id);

  // so now I can look for the tag I want, by name, and that will give me its file-side handle.

  // then I should be able to map it locally, and that should cause
  // the memory-side tag definition to appear at execute time.

  /* Now we can tell damsel to execute everything. */

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, &req);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  int ix, iy;
  for(ix=0; ix<5; ix++) {
    for(iy=0; iy<3; iy++) {
printf("Comparing read value %d with target value %d\n", data[ix][iy], targetdata[ix][iy]);
      if(targetdata[ix][iy] != data[ix][iy]) {
printf("fail!\n");
        return 1;
      }
    }
  }

  DMSLmodel_close(model);
  DMSLcontainer_close(tag_collection_id);
  DMSLcontainer_close(vertex_collection_id);

  linkedlist_free(filetags_ll, free_filetags_container);

  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

