#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "mpi.h"
#include "damsel.h"
#include "damsel-internal.h"
#include "hdf5.h"
#include "damsel-sugar.h"

#include <float.h>

#define DIM1 5
#define NMEMBS 3
#define NAME_LEN 26
#define H5FILE_NAME "output/test-type-struct.h5"

typedef struct {
  char    name[NAME_LEN];
  int32_t population;
  double  location[3];
} city_t;

int double_equiv(double a, double b) {
    if (a > b){
	return ( (a-b) <= DBL_EPSILON);
    } else {
	return ( (b-a) <= DBL_EPSILON);
    }
}
int main(int argc, char **argv) {
  int i;
  city_t refdata[DIM1];
  city_t storedata[DIM1];
  char VAR_NAMES_REF[NMEMBS][NAME_LEN] = { "City Name", "Population", "(Lat, Lon, Alt)"};
  
  damsel_model model, fileside_model;
  damsel_model_internal *model_internal;
  damsel_data mem, file;
  int container_cnt, tag_cnt;
  int handle_counter = 1000;
  damsel_tag_buf_type *all_tags;
  int nitems;

  /* set memory data buffer and values */
  for (i=0; i<DIM1; i++) {
    sprintf(refdata[i].name, "city -- %d", i);
    refdata[i].population = 50000+i;
    refdata[i].location[0] = 87.65+i;
    refdata[i].location[1] = 41.85-i;
    refdata[i].location[2] = 700.3+10*i;

    storedata[i].location[0] = 0xDEADBEEF;
    storedata[i].location[1] = 0xBEEFFACE;
    storedata[i].location[2] = 0xDEAFDEED;
  }

  printf("%s\n", argv[0]);

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  /* TODO: there's no mechanisim to query damsel types.  In order to read back
   * data stored with a call to DMSLtype_struct_create, we need to also create
   * that type, and in the same order as the type was created when written.
   * 
   * This is d-u-m-b-dumb but we have bigger fish to fry */


  char *names[NMEMBS] = { "City Name", "Population", "(Lat, Lon, Alt)"};
  damsel_data_type types[NMEMBS];
  size_t name_max = NAME_LEN;
  size_t dim_location = 3;
  types[0] = DMSLtype_array_create(DAMSEL_DATA_TYPE_BYTES, 1, &name_max);
  types[1] = DAMSEL_DATA_TYPE_INTEGER; 
  types[2] = DMSLtype_array_create(DAMSEL_DATA_TYPE_DOUBLE, 1, &dim_location);

  size_t offsets[NMEMBS] = {HOFFSET(city_t, name), HOFFSET(city_t, population), 
      HOFFSET(city_t,location)};

  /* we'll never actually use geo_location, but need to tell the libray about
   * this type.  "unused variable" warnings are expected. */
  damsel_data_type geo_location = DMSLtype_struct_create(NMEMBS, names, types, 
      offsets, sizeof(city_t));
  assert (geo_location != DAMSEL_DATA_TYPE_INVALID);


  DMSL_CHECK(DMSLmodel_attach(model, H5FILE_NAME, MPI_COMM_WORLD, tb));
  DMSLtrait_release(tb);

  DMSL_CHECK(DMSLmodel_populate(model));

  model_internal= (damsel_model_internal*)model;
  fileside_model = (damsel_model) model_internal->attached_file;

  DMSLmodel_get_tuple_count(model, &container_cnt, &tag_cnt);
  assert(tag_cnt == 1);
  all_tags = calloc(tag_cnt, sizeof(damsel_tag_buf_type));
  DMSL_CHECK(DMSLmodel_get_tag_infos(model, all_tags));

  assert(all_tags[0].tag_datatype == geo_location);
  err = DMSLselect_tag_by_name(model, all_tags[0].name, &file.tag_handle);
  err = DMSLcontainer_create_sequence(fileside_model,
	  file.tag_handle, 1,1, &file.tag_container);

  mem.tag_handle = handle_counter++;
  err = DMSLcontainer_create_sequence(model,
	  mem.tag_handle, 1,1, &mem.tag_container);

  DMSL_CHECK(DMSLmodel_map_handles(mem.tag_container, file.tag_container));

  file.container = DMSLselect_handles_with_values(model, file.tag_handle);

  nitems = DMSLcontainer_count(file.container);
  err = DMSLcontainer_create_sequence(model, mem.container_id, nitems, 1, &mem.container);

  DMSL_CHECK(DMSLmodel_map_handles(mem.container, file.container));
  DMSL_CHECK(DMSLmodel_map_tag(storedata, mem.container, &(mem.tag_handle) ));

  DMSLmodel_transfer_sync(model, DAMSEL_TRANSFER_TYPE_READ);

  DMSLmodel_close(model);
  DMSLcontainer_close(file.container);
  DMSLcontainer_close(mem.container);

  DMSLlib_finalize(lib);

  MPI_Finalize();

  int nerrors=0;
  for (i=0; i<DIM1; i++) {
      if ( !double_equiv(storedata[i].location[0], refdata[i].location[0]) ) {
	  fprintf(stderr, "%dth lon: expected %f, got %f\n", i, 
		  refdata[i].location[0], storedata[i].location[0]);
	  nerrors++;
      }
      if ( !double_equiv(storedata[i].location[1], refdata[i].location[1]) ) {
	  fprintf(stderr, "%dth lat: expected %f, got %f\n", i, 
		  refdata[i].location[1], storedata[i].location[1]);
	  nerrors++;
      }
      if ( !double_equiv(storedata[i].location[2], refdata[i].location[2]) ) {
	  fprintf(stderr, "%dth alt: expected %f, got %f\n", i, 
		  refdata[i].location[2], storedata[i].location[2]);
	  nerrors++;
      }
  }
  return -nerrors;
}


