#include <assert.h>
#include <stdio.h>
#include <unistd.h>

#include "mpi.h"
#include "damsel.h"

#define DIM1 5
#define FILENAME "output/test-type-vector.h5"


int main(int argc, char **argv) {
  int i, j;
  damsel_vector_t data[DIM1];
  
  /* set memory data buffer and values */
  for (i=0; i<DIM1; i++) {
    data[i].len = i+1;
	data[i].p   = calloc(data[i].len, sizeof(short int));
	for (j=0; j<data[i].len; j++)
	  ((short int *)data[i].p)[j] = i*10+j;
  }

  printf("%s\n", argv[0]);
#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  /* Create andatatype to use for the data. purposely make the datype small 
     to test if damsel will recreate a datatype of the right size (DIM1)  */
  damsel_data_type type_id = DMSLtype_vector_create(DAMSEL_DATA_TYPE_SHORTINT, 1);
 
  /* define a tag which will store our data. the tag will be
     named 'somedata' */

  damsel_handle data_tag_handle = 0x10203;
  DMSLtag_define(model, &data_tag_handle, type_id, "somedata");

  /* define a collection of the handles on which we will store data. we
     will store the data on handles 20000 .. upwards. */

  /* we'll create a sequence handle collection with 5 handles in it. */

  damsel_container vertex_collection_id;
  err = DMSLcontainer_create_sequence(model, 0x20000, 5, 0x1000, &vertex_collection_id);

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_collection_id, &data_tag_handle);

  /* at this point, the in-memory store knows about all the data that
     we have. next we will write the data out to disk.
     we need to invent some handles for the file store too. They can overlap
     with memstore handles, though here I make sure that they do not.
  */

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently ignore any error
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_map_handles_inventing_file_handles(vertex_collection_id);

  /* we also need to map the handles for the tags which contain values
     that we want to store.
     This happens by creating a collection containing the tag to map, and
     mapping that into the file store. */

  damsel_container tag_collection_id;
  err = DMSLcontainer_create_sequence(model,
    data_tag_handle, 1, 1, &tag_collection_id);

  DMSLmodel_map_handles_inventing_file_handles(tag_collection_id);

  /* Now we can tell damsel to execute everything. */

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  DMSLmodel_close(model);
  DMSLcontainer_close(vertex_collection_id);
  DMSLcontainer_close(tag_collection_id);

  DMSLlib_finalize(lib);

  MPI_Finalize();
  
  /* free data buff */
  for (i=0; i<DIM1; i++) {
	if (data[i].p)
	  free (data[i].p);
  }
  
  return 0;
}


