/* tests reading into a sequence */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <unistd.h>
#include <memory.h>

#include "damsel.h"
#include "damsel-internal.h"
#include "damsel-debug.h"

#define SIZE 130

int main(int argc, char **argv) {

  // we'll read the data into data[] and then check that
  // it matches up with target data check that
  // it matches up with target data
  short int data[SIZE];

  printf("%s\n", argv[0]);
  if(argc!=2) {
    printf("specify filename of .h5 dump file to check\n");
    return 2;
  }

#ifndef DAMSEL_SINGLETHREAD
  int res;
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);
  damsel_model_internal *ms = (damsel_model_internal *)model;
  assert(ms != NULL);
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "READ");

  DMSLmodel_attach(model, argv[1], MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);
  DMSLmodel_populate(model);

  damsel_handle data_tag_handle = 0x31000;

  /* we'll create a sequence handle collection with 5 handles in it. */

  damsel_container vertex_collection_id;
  err = DMSLcontainer_create_sequence(model, 0x80900, SIZE, 0x5, &vertex_collection_id);

  assert(vertex_collection_id != NULL);

  damsel_container tag_collection_id;
  err = DMSLcontainer_create_sequence(model, 
    data_tag_handle, 1, 1, &tag_collection_id);

  assert(tag_collection_id != NULL);

  damsel_handle fileside_data_tag_handle;
  err = DMSLselect_tag_by_name(model, "somedata", &fileside_data_tag_handle);
  damsel_container file_tag_collection_id;
  err = DMSLcontainer_create_sequence((damsel_model)ms->attached_file, fileside_data_tag_handle, 1, 1, &file_tag_collection_id);
  DMSLmodel_map_handles(tag_collection_id, file_tag_collection_id);

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_collection_id, &data_tag_handle);

  // before we can read in collections, we need to read in containers, so that
  // they can be referred to in the collection structures.
  // this requirement should not stick around.

  int container_count;
  DMSLmodel_get_tuple_count(model, &container_count, NULL);

  damsel_container_buf_type *containerbuf = malloc(container_count * sizeof(damsel_container_buf_type));

  DMSLmodel_get_container_infos(model, containerbuf);

  free(containerbuf);

  int collection_count;
  collection_count = DMSLmodel_get_collection_count(model);
  assert(collection_count == 1);

  damsel_collection_buf_type *collection_buf = malloc(collection_count * sizeof(damsel_collection_buf_type));

  DMSLmodel_get_collection_infos(model, collection_buf);
  damsel_container filestore_vertex_collection_id = collection_buf[0].contents;

  DMSLmodel_close_collection_infos(model, collection_buf);
  free(collection_buf);

  int flength = DMSLcontainer_count(filestore_vertex_collection_id);
  printf("DMSLcontainer_count returned flength = %d\n", flength);
  assert(flength == SIZE);

  DMSLmodel_map_handles(vertex_collection_id, filestore_vertex_collection_id);

  damsel_request_t *req = malloc(sizeof(damsel_request_t));
  assert_valid_model_id(model);
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_READ, req);
  assert_valid_model_id(model);
  printf("MPI request in main thread is %d\n", req->mpi);
  assert_valid_model_id(model);
  damsel_status_t status;
  assert_valid_model_id(model);
  DMSLmodel_wait(*req, &status);
  free(req);
  assert_valid_model_id(model);

  int ix;
  for(ix=0; ix<SIZE; ix++) {
    int i2x = ix / 2;
    int i2r = ix % 2;
    int sw_ix = i2x * 2 + ( 1 - i2r);

    int tv= sw_ix % 77; // must match with expression defined in test-vector-big-in-file.c
printf("Comparing read value %d with target value %d at ix %d\n", data[ix], tv, ix);
    if(tv != data[ix]) {
      printf("fail!\n");
      return 1;
    }
  }

  assert_valid_model_id(model);
  DMSLmodel_close(model);
  DMSLcontainer_close(vertex_collection_id);
  DMSLcontainer_close(tag_collection_id);
  DMSLcontainer_close(file_tag_collection_id);

  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

