/* test create_sequence and writing to disk */

#include <assert.h>
#include <stdio.h>
#include <unistd.h>

#include "mpi.h"
#include "damsel.h"

#define FILENAME "output/test-vector-big-in-file.h5"

// this must be an even number for the test to work
// (because the test code wants to swap every pair of handles, so they must
//  be pairable)
#define SIZE 130

int main(int argc, char **argv) {
#ifndef DAMSEL_SINGLETHREAD
  int res;
#endif

  printf("%s\n", argv[0]);

#ifndef DAMSEL_SINGLETHREAD
  MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &res);
  assert(res == MPI_THREAD_MULTIPLE);
#else
  MPI_Init(&argc, &argv);
#endif

  damsel_library lib;
  damsel_err_t err = DMSLlib_init(&lib);

  damsel_model model;
  err = DMSLmodel_create(DAMSEL_HANDLE_TYPE_HANDLE64, &model);
  assert(model != NULL);

  /* define a tag which will store our data. the tag will be
     named 'somedata' */

  damsel_handle data_tag_handle = 0x10203;
  DMSLtag_define(model, &data_tag_handle, DAMSEL_DATA_TYPE_SHORTINT, "somedata");

  /* we'll create a sequence handle collection with SIZE handles in it. */
;
  const damsel_handle base = 0x5000;
  const size_t step = 0x10;
  damsel_container vertex_container_id;
  err = DMSLcontainer_create_sequence(model,
    base, SIZE, step, &vertex_container_id);

  short int *data = malloc(SIZE * sizeof(short int));
  int ix;
  for(ix=0; ix<SIZE; ix++) {
    data[ix] = ix % 77;
  }

  /* now set the values of the somedata tag on the handles */
  DMSLmodel_map_tag(data, vertex_container_id, &data_tag_handle);

  // now make a collection of the handles, rearranged
  damsel_handle handles_rearranged[SIZE];
  for(ix=0; ix<SIZE; ix+=2) {
    handles_rearranged[ix] = base + (ix + 1) * step;
    handles_rearranged[ix+1] = base + ix * step;
  }

  damsel_container rearranged_container;
  err = DMSLcontainer_create_vector(model, handles_rearranged, SIZE, &rearranged_container);

  /* at this point, the in-memory store knows about all the data that
     we have. next we will write the data out to disk.
     we need to invent some handles for the file store too. They can overlap
     with memstore handles, though here I make sure that they do not.
  */

  unlink(FILENAME); // maybe it doesn't exist, so we'll just silently ignore any error
  damsel_trait_bundle tb;
  err = DMSLtrait_create(&tb);
  DMSLtrait_put(tb, "mode", "WRITE");

  DMSLmodel_attach(model, FILENAME, MPI_COMM_WORLD, tb);
  DMSLtrait_release(tb);

  DMSLmodel_map_handles_inventing_file_handles(vertex_container_id);

  /* we also need to map the handles for the tags which contain values
     that we want to store.
     This happens by creating a collection containing the tag to map, and
     mapping that into the file store. */

  DMSLmodel_map_single_handle_inventing_file_handle(model, data_tag_handle);

  /* now lets make a collection with this in it */

  damsel_handle collection_handle = 0x10204;

  DMSLcoll_create(model, &collection_handle, rearranged_container, DAMSEL_HANDLE_COLLECTION_TYPE_VECTOR);
  DMSLmodel_map_single_handle_inventing_file_handle(model, collection_handle);

  /* Now we can tell damsel to execute everything. */

  damsel_request_t req;
  DMSLmodel_transfer_async(model, DAMSEL_TRANSFER_TYPE_WRITE, &req);
  damsel_status_t status;
  DMSLmodel_wait(req, &status);

  DMSLmodel_close(model);
  free(data);
  DMSLcontainer_close(rearranged_container);
  DMSLcontainer_close(vertex_container_id);

  DMSLlib_finalize(lib);

  MPI_Finalize();
  return 0;
}

